import enum

class TypeIdentity(enum.Enum):
    INTEGER = 1
    ENUM = 2
    CONSTS = 3
    STRUCT = 4
    ARRAY = 5
    STRING = 6

class Type:
    def __init__(self, name, identity, fixed_size = None, dynamic = False, subtype = None, signed = False, n_elements = None, attributes = {}):
        self.name = name
        self.identity = identity
        self.fixed_size = fixed_size
        self.dynamic = dynamic
        self.subtype = subtype
        self.n_elements = None
        self.signed = None
        self.attributes = attributes

        if self.identity is TypeIdentity.ARRAY and n_elements:
            self.n_elements = n_elements

        if self.identity is TypeIdentity.INTEGER:
            self.signed = signed

    def __repr__(self):
        return f'{self.name}'

class TypeRegistry:
    def __init__(self):
        self.types = {
            'int8': Type('int8', TypeIdentity.INTEGER, fixed_size = 1, signed = True),
            'int16': Type('int16', TypeIdentity.INTEGER, fixed_size = 2, signed = True),
            'int32': Type('int32', TypeIdentity.INTEGER, fixed_size = 4, signed = True),
            'int64': Type('int64', TypeIdentity.INTEGER, fixed_size = 8, signed = True),

            'uint8': Type('uint8', TypeIdentity.INTEGER, fixed_size = 1, signed = False),
            'uint16': Type('uint16', TypeIdentity.INTEGER, fixed_size = 2, signed = False),
            'uint32': Type('uint32', TypeIdentity.INTEGER, fixed_size = 4, signed = False),
            'uint64': Type('uint64', TypeIdentity.INTEGER, fixed_size = 8, signed = False),

            'char': Type('char', TypeIdentity.INTEGER, fixed_size = 1),
            'byte': Type('byte', TypeIdentity.INTEGER, fixed_size = 1, signed = False)
        }

        self.types['string'] = Type('string', TypeIdentity.STRING, dynamic = True, subtype = self.types['char'])

    def get_type(self, name):
        return self.types[name] if name in self.types else None

    def register_type(self, type):
        self.types[type.name] = type

    def is_known_type(self, name):
        return name in self.types

    def parse_type(self, name):
        base, delim, size = name.rpartition('[')

        if delim == '':
            if name not in self.types:
                return None
            else:
                return self.types[name]

        if size[-1] != ']':
            return None

        size = size[:-1]
        i_size = None

        if len(size) > 0:
            try:
                i_size = int(size)
            except ValueError:
                return None

        base_type = self.parse_type(base)
        if not base_type:
            return None

        dynamic = base_type.dynamic or len(size) == 0
        t_size = base_type.fixed_size * i_size if not dynamic else None

        return Type(name, TypeIdentity.ARRAY,
                fixed_size = t_size,
                n_elements = i_size,
                dynamic = dynamic,
                subtype = base_type)



from abc import ABC, abstractmethod
import getpass
import sys, os, pickle
import tempfile
import types
import re
from typing import (
    TypeVar, Type, List, Dict, Iterator, Callable, Union, Optional, Sequence,
    Tuple, Iterable, IO, Any, TYPE_CHECKING, Collection
)
if TYPE_CHECKING:
    from .parsers.lalr_interactive_parser import InteractiveParser
    from .tree import ParseTree
    from .visitors import Transformer
    from typing import Literal
    from .parser_frontends import ParsingFrontend

from .exceptions import ConfigurationError, assert_config, UnexpectedInput
from .utils import Serialize, SerializeMemoizer, FS, logger, TextOrSlice, LarkInput
from .load_grammar import load_grammar, FromPackageLoader, Grammar, verify_used_files, PackageResource, sha256_digest
from .tree import Tree
from .common import LexerConf, ParserConf, _ParserArgType, _LexerArgType

from .lexer import Lexer, BasicLexer, TerminalDef, LexerThread, Token
from .parse_tree_builder import ParseTreeBuilder
from .parser_frontends import _validate_frontend_args, _get_lexer_callbacks, _deserialize_parsing_frontend, _construct_parsing_frontend
from .grammar import Rule


try:
    import regex
    _has_regex = True
except ImportError:
    _has_regex = False


###{standalone


class PostLex(ABC):
    @abstractmethod
    def process(self, stream: Iterator[Token]) -> Iterator[Token]:
        return stream

    always_accept: Iterable[str] = ()

class LarkOptions(Serialize):
    """Specifies the options for Lark

    """

    start: List[str]
    debug: bool
    strict: bool
    transformer: 'Optional[Transformer]'
    propagate_positions: Union[bool, str]
    maybe_placeholders: bool
    cache: Union[bool, str]
    cache_grammar: bool
    regex: bool
    g_regex_flags: int
    keep_all_tokens: bool
    tree_class: Optional[Callable[[str, List], Any]]
    parser: _ParserArgType
    lexer: _LexerArgType
    ambiguity: 'Literal["auto", "resolve", "explicit", "forest"]'
    postlex: Optional[PostLex]
    priority: 'Optional[Literal["auto", "normal", "invert"]]'
    lexer_callbacks: Dict[str, Callable[[Token], Token]]
    use_bytes: bool
    ordered_sets: bool
    edit_terminals: Optional[Callable[[TerminalDef], TerminalDef]]
    import_paths: 'List[Union[str, Callable[[Union[None, str, PackageResource], str], Tuple[str, str]]]]'
    source_path: Optional[str]

    OPTIONS_DOC = r"""
    **===  General Options  ===**

    start
            The start symbol. Either a string, or a list of strings for multiple possible starts (Default: "start")
    debug
            Display debug information and extra warnings. Use only when debugging (Default: ``False``)
            When used with Earley, it generates a forest graph as "sppf.png", if 'dot' is installed.
    strict
            Throw an exception on any potential ambiguity, including shift/reduce conflicts, and regex collisions.
    transformer
            Applies the transformer to every parse tree (equivalent to applying it after the parse, but faster)
    propagate_positions
            Propagates positional attributes into the 'meta' attribute of all tree branches.
            Sets attributes: (line, column, end_line, end_column, start_pos, end_pos,
                              container_line, container_column, container_end_line, container_end_column)
            Accepts ``False``, ``True``, or a callable, which will filter which nodes to ignore when propagating.
    maybe_placeholders
            When ``True``, the ``[]`` operator returns ``None`` when not matched.
            When ``False``,  ``[]`` behaves like the ``?`` operator, and returns no value at all.
            (default= ``True``)
    cache
            Cache the results of the Lark grammar analysis, for x2 to x3 faster loading. LALR only for now.

            - When ``False``, does nothing (default)
            - When ``True``, caches to a temporary file in the local directory
            - When given a string, caches to the path pointed by the string
    cache_grammar
            For use with ``cache`` option. When ``True``, the unanalyzed grammar is also included in the cache.
            Useful for classes that require the ``Lark.grammar`` to be present (e.g. Reconstructor).
            (default= ``False``)
    regex
            When True, uses the ``regex`` module instead of the stdlib ``re``.
    g_regex_flags
            Flags that are applied to all terminals (both regex and strings)
    keep_all_tokens
            Prevent the tree builder from automagically removing "punctuation" tokens (Default: ``False``)
    tree_class
            Lark will produce trees comprised of instances of this class instead of the default ``lark.Tree``.

    **=== Algorithm Options ===**

    parser
            Decides which parser engine to use. Accepts "earley" or "lalr". (Default: "earley").
            (there is also a "cyk" option for legacy)
    lexer
            Decides whether or not to use a lexer stage

            - "auto" (default): Choose for me based on the parser
            - "basic": Use a basic lexer
            - "contextual": Stronger lexer (only works with parser="lalr")
            - "dynamic": Flexible and powerful (only with parser="earley")
            - "dynamic_complete": Same as dynamic, but tries *every* variation of tokenizing possible.
    ambiguity
            Decides how to handle ambiguity in the parse. Only relevant if parser="earley"

            - "resolve": The parser will automatically choose the simplest derivation
              (it chooses consistently: greedy for tokens, non-greedy for rules)
            - "explicit": The parser will return all derivations wrapped in "_ambig" tree nodes (i.e. a forest).
            - "forest": The parser will return the root of the shared packed parse forest.

    **=== Misc. / Domain Specific Options ===**

    postlex
            Lexer post-processing (Default: ``None``) Only works with the basic and contextual lexers.
    priority
            How priorities should be evaluated - "auto", ``None``, "normal", "invert" (Default: "auto")
    lexer_callbacks
            Dictionary of callbacks for the lexer. May alter tokens during lexing. Use with caution.
    use_bytes
            Accept an input of type ``bytes`` instead of ``str``.
    ordered_sets
            Should Earley use ordered-sets to achieve stable output (~10% slower than regular sets. Default: True)
    edit_terminals
            A callback for editing the terminals before parse.
    import_paths
            A List of either paths or loader functions to specify from where grammars are imported
    source_path
            Override the source of from where the grammar was loaded. Useful for relative imports and unconventional grammar loading
    **=== End of Options ===**
    """
    if __doc__:
        __doc__ += OPTIONS_DOC


    # Adding a new option needs to be done in multiple places:
    # - In the dictionary below. This is the primary truth of which options `Lark.__init__` accepts
    # - In the docstring above. It is used both for the docstring of `LarkOptions` and `Lark`, and in readthedocs
    # - As an attribute of `LarkOptions` above
    # - Potentially in `_LOAD_ALLOWED_OPTIONS` below this class, when the option doesn't change how the grammar is loaded
    # - Potentially in `lark.tools.__init__`, if it makes sense, and it can easily be passed as a cmd argument
    _defaults: Dict[str, Any] = {
        'debug': False,
        'strict': False,
        'keep_all_tokens': False,
        'tree_class': None,
        'cache': False,
        'cache_grammar': False,
        'postlex': None,
        'parser': 'earley',
        'lexer': 'auto',
        'transformer': None,
        'start': 'start',
        'priority': 'auto',
        'ambiguity': 'auto',
        'regex': False,
        'propagate_positions': False,
        'lexer_callbacks': {},
        'maybe_placeholders': True,
        'edit_terminals': None,
        'g_regex_flags': 0,
        'use_bytes': False,
        'ordered_sets': True,
        'import_paths': [],
        'source_path': None,
        '_plugins': {},
    }

    def __init__(self, options_dict: Dict[str, Any]) -> None:
        o = dict(options_dict)

        options = {}
        for name, default in self._defaults.items():
            if name in o:
                value = o.pop(name)
                if isinstance(default, bool) and name not in ('cache', 'use_bytes', 'propagate_positions'):
                    value = bool(value)
            else:
                value = default

            options[name] = value

        if isinstance(options['start'], str):
            options['start'] = [options['start']]

        self.__dict__['options'] = options


        assert_config(self.parser, ('earley', 'lalr', 'cyk', None))

        if self.parser == 'earley' and self.transformer:
            raise ConfigurationError('Cannot specify an embedded transformer when using the Earley algorithm. '
                             'Please use your transformer on the resulting parse tree, or use a different algorithm (i.e. LALR)')

        if self.cache_grammar and not self.cache:
            raise ConfigurationError('cache_grammar cannot be set when cache is disabled')

        if o:
            raise ConfigurationError("Unknown options: %s" % o.keys())

    def __getattr__(self, name: str) -> Any:
        try:
            return self.__dict__['options'][name]
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, name: str, value: str) -> None:
        assert_config(name, self.options.keys(), "%r isn't a valid option. Expected one of: %s")
        self.options[name] = value

    def serialize(self, memo = None) -> Dict[str, Any]:
        return self.options

    @classmethod
    def deserialize(cls, data: Dict[str, Any], memo: Dict[int, Union[TerminalDef, Rule]]) -> "LarkOptions":
        return cls(data)


# Options that can be passed to the Lark parser, even when it was loaded from cache/standalone.
# These options are only used outside of `load_grammar`.
_LOAD_ALLOWED_OPTIONS = {'postlex', 'transformer', 'lexer_callbacks', 'use_bytes', 'debug', 'g_regex_flags', 'regex', 'propagate_positions', 'tree_class', '_plugins'}

_VALID_PRIORITY_OPTIONS = ('auto', 'normal', 'invert', None)
_VALID_AMBIGUITY_OPTIONS = ('auto', 'resolve', 'explicit', 'forest')


_T = TypeVar('_T', bound="Lark")

class Lark(Serialize):
    """Main interface for the library.

    It's mostly a thin wrapper for the many different parsers, and for the tree constructor.

    Parameters:
        grammar: a string or file-object containing the grammar spec (using Lark's ebnf syntax)
        options: a dictionary controlling various aspects of Lark.

    Example:
        >>> Lark(r'''start: "foo" ''')
        Lark(...)
    """

    source_path: str
    source_grammar: str
    grammar: 'Grammar'
    options: LarkOptions
    lexer: Lexer
    parser: 'ParsingFrontend'
    terminals: Collection[TerminalDef]

    __serialize_fields__ = ['parser', 'rules', 'options']

    def __init__(self, grammar: 'Union[Grammar, str, IO[str]]', **options) -> None:
        self.options = LarkOptions(options)
        re_module: types.ModuleType

        # Update which fields are serialized
        if self.options.cache_grammar:
            self.__serialize_fields__ = self.__serialize_fields__ + ['grammar']

        # Set regex or re module
        use_regex = self.options.regex
        if use_regex:
            if _has_regex:
                re_module = regex
            else:
                raise ImportError('`regex` module must be installed if calling `Lark(regex=True)`.')
        else:
            re_module = re

        # Some, but not all file-like objects have a 'name' attribute
        if self.options.source_path is None:
            try:
                self.source_path = grammar.name  # type: ignore[union-attr]
            except AttributeError:
                self.source_path = '<string>'
        else:
            self.source_path = self.options.source_path

        # Drain file-like objects to get their contents
        try:
            read = grammar.read  # type: ignore[union-attr]
        except AttributeError:
            pass
        else:
            grammar = read()

        cache_fn = None
        cache_sha256 = None
        if isinstance(grammar, str):
            self.source_grammar = grammar
            if self.options.use_bytes:
                if not grammar.isascii():
                    raise ConfigurationError("Grammar must be ascii only, when use_bytes=True")

            if self.options.cache:
                if self.options.parser != 'lalr':
                    raise ConfigurationError("cache only works with parser='lalr' for now")

                unhashable = ('transformer', 'postlex', 'lexer_callbacks', 'edit_terminals', '_plugins')
                options_str = ''.join(k+str(v) for k, v in options.items() if k not in unhashable)
                from . import __version__
                s = grammar + options_str + __version__ + str(sys.version_info[:2])
                cache_sha256 = sha256_digest(s)

                if isinstance(self.options.cache, str):
                    cache_fn = self.options.cache
                else:
                    if self.options.cache is not True:
                        raise ConfigurationError("cache argument must be bool or str")

                    try:
                        username = getpass.getuser()
                    except Exception:
                        # The exception raised may be ImportError or OSError in
                        # the future.  For the cache, we don't care about the
                        # specific reason - we just want a username.
                        username = "unknown"


                    cache_fn = tempfile.gettempdir() + "/.lark_%s_%s_%s_%s_%s.tmp" % (
                        "cache_grammar" if self.options.cache_grammar else "cache", username, cache_sha256, *sys.version_info[:2])

                old_options = self.options
                try:
                    with FS.open(cache_fn, 'rb') as f:
                        logger.debug('Loading grammar from cache: %s', cache_fn)
                        # Remove options that aren't relevant for loading from cache
                        for name in (set(options) - _LOAD_ALLOWED_OPTIONS):
                            del options[name]
                        file_sha256 = f.readline().rstrip(b'\n')
                        cached_used_files = pickle.load(f)
                        if file_sha256 == cache_sha256.encode('utf8') and verify_used_files(cached_used_files):
                            cached_parser_data = pickle.load(f)
                            self._load(cached_parser_data, **options)
                            return
                except FileNotFoundError:
                    # The cache file doesn't exist; parse and compose the grammar as normal
                    pass
                except Exception: # We should probably narrow done which errors we catch here.
                    logger.exception("Failed to load Lark from cache: %r. We will try to carry on.", cache_fn)

                    # In theory, the Lark instance might have been messed up by the call to `_load`.
                    # In practice the only relevant thing that might have been overwritten should be `options`
                    self.options = old_options


            # Parse the grammar file and compose the grammars
            self.grammar, used_files = load_grammar(grammar, self.source_path, self.options.import_paths, self.options.keep_all_tokens)
        else:
            assert isinstance(grammar, Grammar)
            self.grammar = grammar


        if self.options.lexer == 'auto':
            if self.options.parser == 'lalr':
                self.options.lexer = 'contextual'
            elif self.options.parser == 'earley':
                if self.options.postlex is not None:
                    logger.info("postlex can't be used with the dynamic lexer, so we use 'basic' instead. "
                                "Consider using lalr with contextual instead of earley")
                    self.options.lexer = 'basic'
                else:
                    self.options.lexer = 'dynamic'
            elif self.options.parser == 'cyk':
                self.options.lexer = 'basic'
            else:
                assert False, self.options.parser
        lexer = self.options.lexer
        if isinstance(lexer, type):
            assert issubclass(lexer, Lexer)     # XXX Is this really important? Maybe just ensure interface compliance
        else:
            assert_config(lexer, ('basic', 'contextual', 'dynamic', 'dynamic_complete'))
            if self.options.postlex is not None and 'dynamic' in lexer:
                raise ConfigurationError("Can't use postlex with a dynamic lexer. Use basic or contextual instead")

        if self.options.ambiguity == 'auto':
            if self.options.parser == 'earley':
                self.options.ambiguity = 'resolve'
        else:
            assert_config(self.options.parser, ('earley', 'cyk'), "%r doesn't support disambiguation. Use one of these parsers instead: %s")

        if self.options.priority == 'auto':
            self.options.priority = 'normal'

        if self.options.priority not in _VALID_PRIORITY_OPTIONS:
            raise ConfigurationError("invalid priority option: %r. Must be one of %r" % (self.options.priority, _VALID_PRIORITY_OPTIONS))
        if self.options.ambiguity not in _VALID_AMBIGUITY_OPTIONS:
            raise ConfigurationError("invalid ambiguity option: %r. Must be one of %r" % (self.options.ambiguity, _VALID_AMBIGUITY_OPTIONS))

        if self.options.parser is None:
            terminals_to_keep = '*'     # For lexer-only mode, keep all terminals
        elif self.options.postlex is not None:
            terminals_to_keep = set(self.options.postlex.always_accept)
        else:
            terminals_to_keep = set()

        # Compile the EBNF grammar into BNF
        self.terminals, self.rules, self.ignore_tokens = self.grammar.compile(self.options.start, terminals_to_keep)

        if self.options.edit_terminals:
            for t in self.terminals:
                self.options.edit_terminals(t)

        self._terminals_dict = {t.name: t for t in self.terminals}

        # If the user asked to invert the priorities, negate them all here.
        if self.options.priority == 'invert':
            for rule in self.rules:
                if rule.options.priority is not None:
                    rule.options.priority = -rule.options.priority
            for term in self.terminals:
                term.priority = -term.priority
        # Else, if the user asked to disable priorities, strip them from the
        # rules and terminals. This allows the Earley parsers to skip an extra forest walk
        # for improved performance, if you don't need them (or didn't specify any).
        elif self.options.priority is None:
            for rule in self.rules:
                if rule.options.priority is not None:
                    rule.options.priority = None
            for term in self.terminals:
                term.priority = 0

        # TODO Deprecate lexer_callbacks?
        self.lexer_conf = LexerConf(
                self.terminals, re_module, self.ignore_tokens, self.options.postlex,
                self.options.lexer_callbacks, self.options.g_regex_flags, use_bytes=self.options.use_bytes, strict=self.options.strict
            )

        if self.options.parser:
            self.parser = self._build_parser()
        elif lexer:
            self.lexer = self._build_lexer()

        if cache_fn:
            logger.debug('Saving grammar to cache: %s', cache_fn)
            try:
                with FS.open(cache_fn, 'wb') as f:
                    assert cache_sha256 is not None
                    f.write(cache_sha256.encode('utf8') + b'\n')
                    pickle.dump(used_files, f)
                    self.save(f, _LOAD_ALLOWED_OPTIONS)
            except IOError as e:
                logger.exception("Failed to save Lark to cache: %r.", cache_fn, e)

    if __doc__:
        __doc__ += "\n\n" + LarkOptions.OPTIONS_DOC

    def _build_lexer(self, dont_ignore: bool=False) -> BasicLexer:
        lexer_conf = self.lexer_conf
        if dont_ignore:
            from copy import copy
            lexer_conf = copy(lexer_conf)
            lexer_conf.ignore = ()
        return BasicLexer(lexer_conf)

    def _prepare_callbacks(self) -> None:
        self._callbacks = {}
        # we don't need these callbacks if we aren't building a tree
        if self.options.ambiguity != 'forest':
            self._parse_tree_builder = ParseTreeBuilder(
                    self.rules,
                    self.options.tree_class or Tree,
                    self.options.propagate_positions,
                    self.options.parser != 'lalr' and self.options.ambiguity == 'explicit',
                    self.options.maybe_placeholders
                )
            self._callbacks = self._parse_tree_builder.create_callback(self.options.transformer)
        self._callbacks.update(_get_lexer_callbacks(self.options.transformer, self.terminals))

    def _build_parser(self) -> "ParsingFrontend":
        self._prepare_callbacks()
        _validate_frontend_args(self.options.parser, self.options.lexer)
        parser_conf = ParserConf(self.rules, self._callbacks, self.options.start)
        return _construct_parsing_frontend(
            self.options.parser,
            self.options.lexer,
            self.lexer_conf,
            parser_conf,
            options=self.options
        )

    def save(self, f, exclude_options: Collection[str] = ()) -> None:
        """Saves the instance into the given file object

        Useful for caching and multiprocessing.
        """
        if self.options.parser != 'lalr':
            raise NotImplementedError("Lark.save() is only implemented for the LALR(1) parser.")
        data, m = self.memo_serialize([TerminalDef, Rule])
        if exclude_options:
            data["options"] = {n: v for n, v in data["options"].items() if n not in exclude_options}
        pickle.dump({'data': data, 'memo': m}, f, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls: Type[_T], f) -> _T:
        """Loads an instance from the given file object

        Useful for caching and multiprocessing.
        """
        inst = cls.__new__(cls)
        return inst._load(f)

    def _deserialize_lexer_conf(self, data: Dict[str, Any], memo: Dict[int, Union[TerminalDef, Rule]], options: LarkOptions) -> LexerConf:
        lexer_conf = LexerConf.deserialize(data['lexer_conf'], memo)
        lexer_conf.callbacks = options.lexer_callbacks or {}
        lexer_conf.re_module = regex if options.regex else re
        lexer_conf.use_bytes = options.use_bytes
        lexer_conf.g_regex_flags = options.g_regex_flags
        lexer_conf.skip_validation = True
        lexer_conf.postlex = options.postlex
        return lexer_conf

    def _load(self: _T, f: Any, **kwargs) -> _T:
        if isinstance(f, dict):
            d = f
        else:
            d = pickle.load(f)
        memo_json = d['memo']
        data = d['data']

        assert memo_json
        memo = SerializeMemoizer.deserialize(memo_json, {'Rule': Rule, 'TerminalDef': TerminalDef}, {})
        if 'grammar' in data:
            self.grammar = Grammar.deserialize(data['grammar'], memo)
        options = dict(data['options'])
        if (set(kwargs) - _LOAD_ALLOWED_OPTIONS) & set(LarkOptions._defaults):
            raise ConfigurationError("Some options are not allowed when loading a Parser: {}"
                             .format(set(kwargs) - _LOAD_ALLOWED_OPTIONS))
        options.update(kwargs)
        self.options = LarkOptions.deserialize(options, memo)
        self.rules = [Rule.deserialize(r, memo) for r in data['rules']]
        self.source_path = '<deserialized>'
        _validate_frontend_args(self.options.parser, self.options.lexer)
        self.lexer_conf = self._deserialize_lexer_conf(data['parser'], memo, self.options)
        self.terminals = self.lexer_conf.terminals
        self._prepare_callbacks()
        self._terminals_dict = {t.name: t for t in self.terminals}
        self.parser = _deserialize_parsing_frontend(
            data['parser'],
            memo,
            self.lexer_conf,
            self._callbacks,
            self.options,  # Not all, but multiple attributes are used
        )
        return self

    @classmethod
    def _load_from_dict(cls, data, memo, **kwargs):
        inst = cls.__new__(cls)
        return inst._load({'data': data, 'memo': memo}, **kwargs)

    @classmethod
    def open(cls: Type[_T], grammar_filename: str, rel_to: Optional[str]=None, **options) -> _T:
        """Create an instance of Lark with the grammar given by its filename

        If ``rel_to`` is provided, the function will find the grammar filename in relation to it.

        Example:

            >>> Lark.open("grammar_file.lark", rel_to=__file__, parser="lalr")
            Lark(...)

        """
        if rel_to:
            basepath = os.path.dirname(rel_to)
            grammar_filename = os.path.join(basepath, grammar_filename)
        with open(grammar_filename, encoding='utf8') as f:
            return cls(f, **options)

    @classmethod
    def open_from_package(cls: Type[_T], package: str, grammar_path: str, search_paths: 'Sequence[str]'=[""], **options) -> _T:
        """Create an instance of Lark with the grammar loaded from within the package `package`.
        This allows grammar loading from zipapps.

        Imports in the grammar will use the `package` and `search_paths` provided, through `FromPackageLoader`

        Example:

            Lark.open_from_package(__name__, "example.lark", ("grammars",), parser=...)
        """
        package_loader = FromPackageLoader(package, search_paths)
        full_path, text = package_loader(None, grammar_path)
        options.setdefault('source_path', full_path)
        options.setdefault('import_paths', [])
        options['import_paths'].append(package_loader)
        return cls(text, **options)

    def __repr__(self):
        return 'Lark(open(%r), parser=%r, lexer=%r, ...)' % (self.source_path, self.options.parser, self.options.lexer)


    def lex(self, text: TextOrSlice, dont_ignore: bool=False) -> Iterator[Token]:
        """Only lex (and postlex) the text, without parsing it. Only relevant when lexer='basic'

        When dont_ignore=True, the lexer will return all tokens, even those marked for %ignore.

        :raises UnexpectedCharacters: In case the lexer cannot find a suitable match.
        """
        lexer: Lexer
        if not hasattr(self, 'lexer') or dont_ignore:
            lexer = self._build_lexer(dont_ignore)
        else:
            lexer = self.lexer
        lexer_thread = LexerThread.from_text(lexer, text)
        stream = lexer_thread.lex(None)
        if self.options.postlex:
            return self.options.postlex.process(stream)
        return stream

    def get_terminal(self, name: str) -> TerminalDef:
        """Get information about a terminal"""
        return self._terminals_dict[name]

    def parse_interactive(self, text: Optional[LarkInput]=None, start: Optional[str]=None) -> 'InteractiveParser':
        """Start an interactive parsing session. Only works when parser='lalr'.

        Parameters:
            text (LarkInput, optional): Text to be parsed. Required for ``resume_parse()``.
            start (str, optional): Start symbol

        Returns:
            A new InteractiveParser instance.

        See Also: ``Lark.parse()``
        """
        return self.parser.parse_interactive(text, start=start)

    def parse(self, text: LarkInput, start: Optional[str]=None, on_error: 'Optional[Callable[[UnexpectedInput], bool]]'=None) -> 'ParseTree':
        """Parse the given text, according to the options provided.

        Parameters:
            text (LarkInput): Text to be parsed, as `str` or `bytes`.
                TextSlice may also be used, but only when lexer='basic' or 'contextual'.
                If Lark was created with a custom lexer, this may be an object of any type.
            start (str, optional): Required if Lark was given multiple possible start symbols (using the start option).
            on_error (function, optional): if provided, will be called on UnexpectedInput error,
                with the exception as its argument. Return true to resume parsing, or false to raise the exception.
                LALR only. See examples/advanced/error_handling.py for an example of how to use on_error.

        Returns:
            If a transformer is supplied to ``__init__``, returns whatever is the
            result of the transformation. Otherwise, returns a Tree instance.

        :raises UnexpectedInput: On a parse error, one of these sub-exceptions will rise:
                ``UnexpectedCharacters``, ``UnexpectedToken``, or ``UnexpectedEOF``.
                For convenience, these sub-exceptions also inherit from ``ParserError`` and ``LexerError``.

        """
        if on_error is not None and self.options.parser != 'lalr':
            raise NotImplementedError("The on_error option is only implemented for the LALR(1) parser.")
        return self.parser.parse(text, start=start, on_error=on_error)


###}



import sys
from copy import deepcopy

from typing import List, Callable, Iterator, Union, Optional, Generic, TypeVar, TYPE_CHECKING

from .lexer import Token

if TYPE_CHECKING:
    from .lexer import TerminalDef
    try:
        import rich
    except ImportError:
        pass
    from typing import Literal

###{standalone

class Meta:

    empty: bool
    line: int
    column: int
    start_pos: int
    end_line: int
    end_column: int
    end_pos: int
    orig_expansion: 'List[TerminalDef]'
    match_tree: bool

    def __init__(self):
        self.empty = True


_Leaf_T = TypeVar("_Leaf_T")
Branch = Union[_Leaf_T, 'Tree[_Leaf_T]']


class Tree(Generic[_Leaf_T]):
    """The main tree class.

    Creates a new tree, and stores "data" and "children" in attributes of the same name.
    Trees can be hashed and compared.

    Parameters:
        data: The name of the rule or alias
        children: List of matched sub-rules and terminals
        meta: Line & Column numbers (if ``propagate_positions`` is enabled).
            meta attributes: (line, column, end_line, end_column, start_pos, end_pos,
                              container_line, container_column, container_end_line, container_end_column)
            container_* attributes consider all symbols, including those that have been inlined in the tree.
            For example, in the rule 'a: _A B _C', the regular attributes will mark the start and end of B,
            but the container_* attributes will also include _A and _C in the range. However, rules that
            contain 'a' will consider it in full, including _A and _C for all attributes.
    """

    data: str
    children: 'List[Branch[_Leaf_T]]'

    def __init__(self, data: str, children: 'List[Branch[_Leaf_T]]', meta: Optional[Meta]=None) -> None:
        self.data = data
        self.children = children
        self._meta = meta

    @property
    def meta(self) -> Meta:
        if self._meta is None:
            self._meta = Meta()
        return self._meta

    def __repr__(self):
        return 'Tree(%r, %r)' % (self.data, self.children)

    __match_args__ = ("data", "children")

    def _pretty_label(self):
        return self.data

    def _pretty(self, level, indent_str):
        yield f'{indent_str*level}{self._pretty_label()}'
        if len(self.children) == 1 and not isinstance(self.children[0], Tree):
            yield f'\t{self.children[0]}\n'
        else:
            yield '\n'
            for n in self.children:
                if isinstance(n, Tree):
                    yield from n._pretty(level+1, indent_str)
                else:
                    yield f'{indent_str*(level+1)}{n}\n'

    def pretty(self, indent_str: str='  ') -> str:
        """Returns an indented string representation of the tree.

        Great for debugging.
        """
        return ''.join(self._pretty(0, indent_str))

    def __rich__(self, parent:Optional['rich.tree.Tree']=None) -> 'rich.tree.Tree':
        """Returns a tree widget for the 'rich' library.

        Example:
            ::
                from rich import print
                from lark import Tree

                tree = Tree('root', ['node1', 'node2'])
                print(tree)
        """
        return self._rich(parent)

    def _rich(self, parent):
        if parent:
            tree = parent.add(f'[bold]{self.data}[/bold]')
        else:
            import rich.tree
            tree = rich.tree.Tree(self.data)

        for c in self.children:
            if isinstance(c, Tree):
                c._rich(tree)
            else:
                tree.add(f'[green]{c}[/green]')

        return tree

    def __eq__(self, other):
        try:
            return self.data == other.data and self.children == other.children
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self) -> int:
        return hash((self.data, tuple(self.children)))

    def iter_subtrees(self) -> 'Iterator[Tree[_Leaf_T]]':
        """Depth-first iteration.

        Iterates over all the subtrees, never returning to the same node twice (Lark's parse-tree is actually a DAG).
        """
        queue = [self]
        subtrees = dict()
        for subtree in queue:
            subtrees[id(subtree)] = subtree
            queue += [c for c in reversed(subtree.children)
                      if isinstance(c, Tree) and id(c) not in subtrees]

        del queue
        return reversed(list(subtrees.values()))

    def iter_subtrees_topdown(self):
        """Breadth-first iteration.

        Iterates over all the subtrees, return nodes in order like pretty() does.
        """
        stack = [self]
        stack_append = stack.append
        stack_pop = stack.pop
        while stack:
            node = stack_pop()
            if not isinstance(node, Tree):
                continue
            yield node
            for child in reversed(node.children):
                stack_append(child)

    def find_pred(self, pred: 'Callable[[Tree[_Leaf_T]], bool]') -> 'Iterator[Tree[_Leaf_T]]':
        """Returns all nodes of the tree that evaluate pred(node) as true."""
        return filter(pred, self.iter_subtrees())

    def find_data(self, data: str) -> 'Iterator[Tree[_Leaf_T]]':
        """Returns all nodes of the tree whose data equals the given data."""
        return self.find_pred(lambda t: t.data == data)

###}

    def find_token(self, token_type: str) -> Iterator[_Leaf_T]:
        """Returns all tokens whose type equals the given token_type.

        This is a recursive function that will find tokens in all the subtrees.

        Example:
            >>> term_tokens = tree.find_token('TERM')
        """
        return self.scan_values(lambda v: isinstance(v, Token) and v.type == token_type)

    def expand_kids_by_data(self, *data_values):
        """Expand (inline) children with any of the given data values. Returns True if anything changed"""
        changed = False
        for i in range(len(self.children)-1, -1, -1):
            child = self.children[i]
            if isinstance(child, Tree) and child.data in data_values:
                self.children[i:i+1] = child.children
                changed = True
        return changed


    def scan_values(self, pred: 'Callable[[Branch[_Leaf_T]], bool]') -> Iterator[_Leaf_T]:
        """Return all values in the tree that evaluate pred(value) as true.

        This can be used to find all the tokens in the tree.

        Example:
            >>> all_tokens = tree.scan_values(lambda v: isinstance(v, Token))
        """
        for c in self.children:
            if isinstance(c, Tree):
                for t in c.scan_values(pred):
                    yield t
            else:
                if pred(c):
                    yield c

    def __deepcopy__(self, memo):
        return type(self)(self.data, deepcopy(self.children, memo), meta=self._meta)

    def copy(self) -> 'Tree[_Leaf_T]':
        return type(self)(self.data, self.children)

    def set(self, data: str, children: 'List[Branch[_Leaf_T]]') -> None:
        self.data = data
        self.children = children


ParseTree = Tree['Token']


class SlottedTree(Tree):
    __slots__ = 'data', 'children', 'rule', '_meta'


def pydot__tree_to_png(tree: Tree, filename: str, rankdir: 'Literal["TB", "LR", "BT", "RL"]'="LR", **kwargs) -> None:
    graph = pydot__tree_to_graph(tree, rankdir, **kwargs)
    graph.write_png(filename)


def pydot__tree_to_dot(tree: Tree, filename, rankdir="LR", **kwargs):
    graph = pydot__tree_to_graph(tree, rankdir, **kwargs)
    graph.write(filename)


def pydot__tree_to_graph(tree: Tree, rankdir="LR", **kwargs):
    """Creates a colorful image that represents the tree (data+children, without meta)

    Possible values for `rankdir` are "TB", "LR", "BT", "RL", corresponding to
    directed graphs drawn from top to bottom, from left to right, from bottom to
    top, and from right to left, respectively.

    `kwargs` can be any graph attribute (e. g. `dpi=200`). For a list of
    possible attributes, see https://www.graphviz.org/doc/info/attrs.html.
    """

    import pydot  # type: ignore[import-not-found]
    graph = pydot.Dot(graph_type='digraph', rankdir=rankdir, **kwargs)

    i = [0]

    def new_leaf(leaf):
        node = pydot.Node(i[0], label=repr(leaf))
        i[0] += 1
        graph.add_node(node)
        return node

    def _to_pydot(subtree):
        color = hash(subtree.data) & 0xffffff
        color |= 0x808080

        subnodes = [_to_pydot(child) if isinstance(child, Tree) else new_leaf(child)
                    for child in subtree.children]
        node = pydot.Node(i[0], style="filled", fillcolor="#%x" % color, label=subtree.data)
        i[0] += 1
        graph.add_node(node)

        for subnode in subnodes:
            graph.add_edge(pydot.Edge(node, subnode))

        return node

    _to_pydot(tree)
    return graph



from __future__ import annotations

import contextlib
import functools
import logging
import os
from typing import TYPE_CHECKING, cast

from pip._vendor.packaging.utils import canonicalize_name
from pip._vendor.resolvelib import BaseReporter, ResolutionImpossible, ResolutionTooDeep
from pip._vendor.resolvelib import Resolver as RLResolver
from pip._vendor.resolvelib.structs import DirectedGraph

from pip._internal.cache import WheelCache
from pip._internal.exceptions import ResolutionTooDeepError
from pip._internal.index.package_finder import PackageFinder
from pip._internal.operations.prepare import RequirementPreparer
from pip._internal.req.constructors import install_req_extend_extras
from pip._internal.req.req_install import InstallRequirement
from pip._internal.req.req_set import RequirementSet
from pip._internal.resolution.base import BaseResolver, InstallRequirementProvider
from pip._internal.resolution.resolvelib.provider import PipProvider
from pip._internal.resolution.resolvelib.reporter import (
    PipDebuggingReporter,
    PipReporter,
)
from pip._internal.utils.packaging import get_requirement

from .base import Candidate, Requirement
from .factory import Factory

if TYPE_CHECKING:
    from pip._vendor.resolvelib.resolvers import Result as RLResult

    Result = RLResult[Requirement, Candidate, str]


logger = logging.getLogger(__name__)


class Resolver(BaseResolver):
    _allowed_strategies = {"eager", "only-if-needed", "to-satisfy-only"}

    def __init__(
        self,
        preparer: RequirementPreparer,
        finder: PackageFinder,
        wheel_cache: WheelCache | None,
        make_install_req: InstallRequirementProvider,
        use_user_site: bool,
        ignore_dependencies: bool,
        ignore_installed: bool,
        ignore_requires_python: bool,
        force_reinstall: bool,
        upgrade_strategy: str,
        py_version_info: tuple[int, ...] | None = None,
    ):
        super().__init__()
        assert upgrade_strategy in self._allowed_strategies

        self.factory = Factory(
            finder=finder,
            preparer=preparer,
            make_install_req=make_install_req,
            wheel_cache=wheel_cache,
            use_user_site=use_user_site,
            force_reinstall=force_reinstall,
            ignore_installed=ignore_installed,
            ignore_requires_python=ignore_requires_python,
            py_version_info=py_version_info,
        )
        self.ignore_dependencies = ignore_dependencies
        self.upgrade_strategy = upgrade_strategy
        self._result: Result | None = None

    def resolve(
        self, root_reqs: list[InstallRequirement], check_supported_wheels: bool
    ) -> RequirementSet:
        collected = self.factory.collect_root_requirements(root_reqs)
        provider = PipProvider(
            factory=self.factory,
            constraints=collected.constraints,
            ignore_dependencies=self.ignore_dependencies,
            upgrade_strategy=self.upgrade_strategy,
            user_requested=collected.user_requested,
        )
        if "PIP_RESOLVER_DEBUG" in os.environ:
            reporter: BaseReporter[Requirement, Candidate, str] = PipDebuggingReporter()
        else:
            reporter = PipReporter()
        resolver: RLResolver[Requirement, Candidate, str] = RLResolver(
            provider,
            reporter,
        )

        try:
            limit_how_complex_resolution_can_be = 200000
            result = self._result = resolver.resolve(
                collected.requirements, max_rounds=limit_how_complex_resolution_can_be
            )

        except ResolutionImpossible as e:
            error = self.factory.get_installation_error(
                cast("ResolutionImpossible[Requirement, Candidate]", e),
                collected.constraints,
            )
            raise error from e
        except ResolutionTooDeep:
            raise ResolutionTooDeepError from None

        req_set = RequirementSet(check_supported_wheels=check_supported_wheels)
        # process candidates with extras last to ensure their base equivalent is
        # already in the req_set if appropriate.
        # Python's sort is stable so using a binary key function keeps relative order
        # within both subsets.
        for candidate in sorted(
            result.mapping.values(), key=lambda c: c.name != c.project_name
        ):
            ireq = candidate.get_install_requirement()
            if ireq is None:
                if candidate.name != candidate.project_name:
                    # extend existing req's extras
                    with contextlib.suppress(KeyError):
                        req = req_set.get_requirement(candidate.project_name)
                        req_set.add_named_requirement(
                            install_req_extend_extras(
                                req, get_requirement(candidate.name).extras
                            )
                        )
                continue

            # Check if there is already an installation under the same name,
            # and set a flag for later stages to uninstall it, if needed.
            installed_dist = self.factory.get_dist_to_uninstall(candidate)
            if installed_dist is None:
                # There is no existing installation -- nothing to uninstall.
                ireq.should_reinstall = False
            elif self.factory.force_reinstall:
                # The --force-reinstall flag is set -- reinstall.
                ireq.should_reinstall = True
            elif installed_dist.version != candidate.version:
                # The installation is different in version -- reinstall.
                ireq.should_reinstall = True
            elif candidate.is_editable or installed_dist.editable:
                # The incoming distribution is editable, or different in
                # editable-ness to installation -- reinstall.
                ireq.should_reinstall = True
            elif candidate.source_link and candidate.source_link.is_file:
                # The incoming distribution is under file://
                if candidate.source_link.is_wheel:
                    # is a local wheel -- do nothing.
                    logger.info(
                        "%s is already installed with the same version as the "
                        "provided wheel. Use --force-reinstall to force an "
                        "installation of the wheel.",
                        ireq.name,
                    )
                    continue

                # is a local sdist or path -- reinstall
                ireq.should_reinstall = True
            else:
                continue

            link = candidate.source_link
            if link and link.is_yanked:
                # The reason can contain non-ASCII characters, Unicode
                # is required for Python 2.
                msg = (
                    "The candidate selected for download or install is a "
                    "yanked version: {name!r} candidate (version {version} "
                    "at {link})\nReason for being yanked: {reason}"
                ).format(
                    name=candidate.name,
                    version=candidate.version,
                    link=link,
                    reason=link.yanked_reason or "<none given>",
                )
                logger.warning(msg)

            req_set.add_named_requirement(ireq)

        reqs = req_set.all_requirements
        self.factory.preparer.prepare_linked_requirements_more(reqs)
        for req in reqs:
            req.prepared = True
            req.needs_more_preparation = False
        return req_set

    def get_installation_order(
        self, req_set: RequirementSet
    ) -> list[InstallRequirement]:
        """Get order for installation of requirements in RequirementSet.

        The returned list contains a requirement before another that depends on
        it. This helps ensure that the environment is kept consistent as they
        get installed one-by-one.

        The current implementation creates a topological ordering of the
        dependency graph, giving more weight to packages with less
        or no dependencies, while breaking any cycles in the graph at
        arbitrary points. We make no guarantees about where the cycle
        would be broken, other than it *would* be broken.
        """
        assert self._result is not None, "must call resolve() first"

        if not req_set.requirements:
            # Nothing is left to install, so we do not need an order.
            return []

        graph = self._result.graph
        weights = get_topological_weights(graph, set(req_set.requirements.keys()))

        sorted_items = sorted(
            req_set.requirements.items(),
            key=functools.partial(_req_set_item_sorter, weights=weights),
            reverse=True,
        )
        return [ireq for _, ireq in sorted_items]


def get_topological_weights(
    graph: DirectedGraph[str | None], requirement_keys: set[str]
) -> dict[str | None, int]:
    """Assign weights to each node based on how "deep" they are.

    This implementation may change at any point in the future without prior
    notice.

    We first simplify the dependency graph by pruning any leaves and giving them
    the highest weight: a package without any dependencies should be installed
    first. This is done again and again in the same way, giving ever less weight
    to the newly found leaves. The loop stops when no leaves are left: all
    remaining packages have at least one dependency left in the graph.

    Then we continue with the remaining graph, by taking the length for the
    longest path to any node from root, ignoring any paths that contain a single
    node twice (i.e. cycles). This is done through a depth-first search through
    the graph, while keeping track of the path to the node.

    Cycles in the graph result would result in node being revisited while also
    being on its own path. In this case, take no action. This helps ensure we
    don't get stuck in a cycle.

    When assigning weight, the longer path (i.e. larger length) is preferred.

    We are only interested in the weights of packages that are in the
    requirement_keys.
    """
    path: set[str | None] = set()
    weights: dict[str | None, list[int]] = {}

    def visit(node: str | None) -> None:
        if node in path:
            # We hit a cycle, so we'll break it here.
            return

        # The walk is exponential and for pathologically connected graphs (which
        # are the ones most likely to contain cycles in the first place) it can
        # take until the heat-death of the universe. To counter this we limit
        # the number of attempts to visit (i.e. traverse through) any given
        # node. We choose a value here which gives decent enough coverage for
        # fairly well behaved graphs, and still limits the walk complexity to be
        # linear in nature.
        cur_weights = weights.get(node, [])
        if len(cur_weights) >= 5:
            return

        # Time to visit the children!
        path.add(node)
        for child in graph.iter_children(node):
            visit(child)
        path.remove(node)

        if node not in requirement_keys:
            return

        cur_weights.append(len(path))
        weights[node] = cur_weights

    # Simplify the graph, pruning leaves that have no dependencies. This is
    # needed for large graphs (say over 200 packages) because the `visit`
    # function is slower for large/densely connected graphs, taking minutes.
    # See https://github.com/pypa/pip/issues/10557
    # We repeat the pruning step until we have no more leaves to remove.
    while True:
        leaves = set()
        for key in graph:
            if key is None:
                continue
            for _child in graph.iter_children(key):
                # This means we have at least one child
                break
            else:
                # No child.
                leaves.add(key)
        if not leaves:
            # We are done simplifying.
            break
        # Calculate the weight for the leaves.
        weight = len(graph) - 1
        for leaf in leaves:
            if leaf not in requirement_keys:
                continue
            weights[leaf] = [weight]
        # Remove the leaves from the graph, making it simpler.
        for leaf in leaves:
            graph.remove(leaf)

    # Visit the remaining graph, this will only have nodes to handle if the
    # graph had a cycle in it, which the pruning step above could not handle.
    # `None` is guaranteed to be the root node by resolvelib.
    visit(None)

    # Sanity check: all requirement keys should be in the weights,
    # and no other keys should be in the weights.
    difference = set(weights.keys()).difference(requirement_keys)
    assert not difference, difference

    # Now give back all the weights, choosing the largest ones from what we
    # accumulated.
    return {node: max(wgts) for (node, wgts) in weights.items()}


def _req_set_item_sorter(
    item: tuple[str, InstallRequirement],
    weights: dict[str | None, int],
) -> tuple[int, str]:
    """Key function used to sort install requirements for installation.

    Based on the "weight" mapping calculated in ``get_installation_order()``.
    The canonical package name is returned as the second member as a tie-
    breaker to ensure the result is predictable, which is useful in tests.
    """
    name = canonicalize_name(item[0])
    return weights[name], name



from .tokens import *
from .types import *

import hashlib

flatten = lambda l: [item for sublist in l for item in sublist]

class CodeGenerator:
    def __init__(self, units):
        self.units = units
        self.indent_depth = 0
        self.indent = ''
        self.enums = {}
        self.messages = {}
        self.namespace_tag = None
        self.protocols = {}

    def enter_indent(self):
        self.indent_depth += 1
        self.indent = '\t' * self.indent_depth
        return self.indent

    def leave_indent(self):
        self.indent_depth -= 1
        self.indent = '\t' * self.indent_depth
        return self.indent

    def generate_header(self):
        '''
        Generate the output file header
        '''

        out = f'{self.indent}function parse_varint(varint)\n'
        self.enter_indent()
        out += f'{self.indent}local nbytes = 9\n'
        out += f'{self.indent}local ret = 0\n'
        out += f'{self.indent}if varint(0, 1):uint() ~= 0 then\n'
        self.enter_indent()
        out += f'{self.indent}for i=1,8 do\n'
        self.enter_indent()
        out += f'{self.indent}if bit.band(bit.rshift(varint(0, 1):uint(), i - 1), 1) ~= 0 then\n'
        self.enter_indent()
        out += f'{self.indent}nbytes = i\n'
        out += f'{self.indent}break\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        out += f'{self.indent}for i=1,nbytes-1 do\n'
        self.enter_indent()
        out += f'{self.indent}ret = bit.bor(ret, bit.lshift(varint(i, 1):uint(), (i - 1) * 8))\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        out += f'{self.indent}local shift = 0\n'
        out += f'{self.indent}if nbytes < 9 then\n'
        self.enter_indent()
        out += f'{self.indent}shift = 8 - (nbytes % 8)\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        out += f'{self.indent}ret = bit.lshift(ret, shift)\n'
        out += f'{self.indent}ret = bit.bor(ret, bit.rshift(varint(0, 1):uint(), nbytes))\n'
        out += f'{self.indent}return nbytes, ret\n'
        self.leave_indent()
        out += f'{self.indent}end\n\n'
        return out

    def generate_footer(self):
        '''
        Generate the output file footer
        '''

        out = f'{self.indent}bragi_protocol = Proto("bragi", "bragi request")\n\n'
        out += f'{self.indent}protocol_id = ProtoField.uint32("bragi.protocol", "Encapsulated protocol", base.HEX, protocol_mappings)\n'
        out += f'{self.indent}message_id = ProtoField.int32("bragi.message_id", "message ID", base.DEC)\n'
        out += f'{self.indent}tail_size = ProtoField.int32("bragi.tail_size", "tail size", base.DEC)\n'
        out += f'{self.indent}pid = ProtoField.uint32("bragi.pid", "PID", base.DEC)\n'
        out += f'{self.indent}request_of = ProtoField.framenum("bragi.request_of", "Response in", base.NONE, frametype.RESPONSE)\n'
        out += f'{self.indent}response_to = ProtoField.framenum("bragi.reply_to", "Reply to", base.NONE, frametype.REQUEST)\n'
        out += f'{self.indent}convo_time = ProtoField.relative_time("bragi.convo_time", "Handling time")\n\n'
        out += f'{self.indent}bragi_protocol.fields = {{protocol_id, message_id, tail_size, pid, request_of, response_to, convo_time}}\n\n'
        out += f'{self.indent}function bragi_protocol.dissector(tvb, pinfo, tree)\n'
        self.enter_indent()
        out += f'{self.indent}if tvb:len() == 0 then return end\n\n'
        out += f'{self.indent}local subtree = tree:add(bragi_protocol, tvb(0, 40), "Bragi Protocol Data")\n'
        out += f'{self.indent}local proto_id = tvb(0, 4)\n'
        out += f'{self.indent}local pid_val = tvb(4, 4)\n'
        out += f'{self.indent}local req_num = tvb(8, 8)\n'
        out += f'{self.indent}local resp_num = tvb(16, 8)\n'
        out += f'{self.indent}local req_time = tvb(24, 8)\n'
        out += f'{self.indent}local request = tvb(32, 4)\n\n'
        out += f'{self.indent}if protocol_mappings[proto_id:le_uint()] ~= nil then\n'
        self.enter_indent()
        out += f'{self.indent}pinfo.cols.protocol = protocol_mappings[proto_id:le_uint()]\n'
        self.leave_indent()
        out += f'{self.indent}else\n'
        self.enter_indent()
        out += f'{self.indent}pinfo.cols.protocol = "bragi"\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        out += f'{self.indent}subtree:add_le(protocol_id, proto_id)\n'
        out += f'{self.indent}subtree:add_le(pid, pid_val)\n'
        out += f'{self.indent}if req_num:le_uint64():tonumber() == pinfo.number and resp_num:le_uint64():tonumber() ~= 0 then\n'
        self.enter_indent()
        out += f'{self.indent}subtree:add(request_of, resp_num, resp_num:le_uint64():tonumber())\n'
        self.leave_indent()
        out += f'{self.indent}elseif resp_num:le_uint64():tonumber() == pinfo.number and req_num:le_uint64():tonumber() ~= 0 then\n'
        self.enter_indent()
        out += f'{self.indent}subtree:add(response_to, req_num, req_num:le_uint64():tonumber())\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        out += f'{self.indent}if req_time:le_uint64():tonumber() ~= 0 then\n'
        self.enter_indent()
        out += f'{self.indent}subtree:add(convo_time, req_time, NSTime.new(req_time:le_uint64():tonumber() // 1000000000, req_time:le_uint64():tonumber() % 1000000000))\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        out += f'{self.indent}if protocols[proto_id:le_uint()] ~= nil then\n'
        self.enter_indent()
        out += f'{self.indent}local proto = protocols[proto_id:le_uint()]\n'
        out += f'{self.indent}if proto[request:le_uint()] ~= nil then\n'
        self.enter_indent()
        out += f'{self.indent}local msg_ids = protocol_message_ids[proto_id:le_uint()]\n'
        out += f'{self.indent}pinfo.cols.info:set("PID " .. pid_val:le_uint() .. " " .. msg_ids[request:le_uint()])\n'
        out += f'{self.indent}subtree:add_le(message_id, request, request:le_uint(), "message:", msg_ids[request:le_uint()])\n'
        out += f'{self.indent}local prototree = tree:add(proto[request:le_uint()], tvb(40, tvb:len() - 40))\n'
        out += f'{self.indent}proto[request:le_uint()].dissector(tvb(40, tvb:len() - 40):tvb(), pinfo, prototree)\n'
        self.leave_indent()
        out += f'{self.indent}end\n'
        self.leave_indent()
        out += f'{self.indent}end\n\n'
        out += f'{self.indent}subtree:add_le(tail_size, tvb(36, 4))\n'
        self.leave_indent()
        out += 'end\n\n'
        out += 'local tab = DissectorTable.get("wtap_encap")\n'
        out += 'tab:add(wtap.USER0, bragi_protocol)\n'
        return out

    def generate_integer_parsing(self, m: MessageMember, msg: Message | Struct, tree_name: str = 'tree', tvb_name: str = 'tvb') -> str:
        '''
        Generates the code to parse an integer field

        Parameters:
            m (MessageMember): The member to generate the parsing code for
            msg (Message | Struct): The message the member belongs to
            tree_name (str): The name of the wireshark tree to add the field to
            tvb_name (str): The name of the tvb to read the field from

        Returns:
            str: The generated code excluding increasing the offset (this needs to be done by the caller). The varint size is accessible via the local `varint_size`.
        '''
        if self.proto_field_type(m.type) in ('uint8', 'char'):
            out = f'{self.indent}local varint_size = 1\n'
            out += f'{self.indent}local {m.name}_item = {tree_name}:add_le({self.proto_field_member_name(msg, m)}, {tvb_name}(offset, varint_size))\n'
        else:
            out = f'{self.indent}local varint_size, value = parse_varint({tvb_name}(offset))\n'
            if m.format and m.format.value == "ipv4":
                out += f'{self.indent}local {m.name}_item = {tree_name}:add_le({self.proto_field_member_name(msg, m)}, {tvb_name}(offset, varint_size), ByteArray.new(string.format("%08x", value)):tvb():range():ipv4())\n'
            else:
                type_prefix, type_suffix = self.wireshark_lua_type(m.type)
                out += f'{self.indent}local {m.name}_item = {tree_name}:add_le({self.proto_field_member_name(msg, m)}, {tvb_name}(offset, varint_size), {type_prefix}value{type_suffix})\n'

        return out

    def generate_string_parsing(self, m: MessageMember, msg: Message | Struct, tree_name: str = 'tree', tvb_name: str = 'tvb', tvb_offset: str = '0') -> str:
        '''
        Generates the code to parse a string field

        Parameters:
            m (MessageMember): The member to generate the parsing code for
            msg (Message | Struct): The message the member belongs to
            tree_name (str): The name of the wireshark tree to add the field to
            tvb_name (str): The name of the tvb to read the field from
            tvb_offset (str): The offset in the tvb to start reading from

        Returns:
            str: The generated code excluding increasing the offset (this needs to be done by the caller). The varint and string length are accessible via the locals `len_size` and `len`.
        '''
        out = f'{self.indent}local len_size, len = parse_varint({tvb_name}({tvb_offset}))\n'
        out += f'{self.indent}local {m.name}_item = {tree_name}:add({self.proto_field_member_name(msg, m)}, {tvb_name}({tvb_offset} + len_size, len))\n'
        return out

    def generate_array_parsing(self, m, message: Message | Struct, loc: str, tree_name: str = 'tree', tvb_name: str = 'tvb', member_offset: str = 'value_size') -> str:
        '''
        Generates the code to parse an array

        Parameters:
            m (MessageMember): The member to generate the parsing code for
            message (Message | Struct): The message the member belongs to
            loc (str): the location information to use in error messages
            tree_name (str): The name of the wireshark tree to add the field to
            tvb_name (str): The name of the tvb to read the field from
            member_offset (str): The variable keeping track of the current offset into the tvb; this variable is increased by the size of the array

        Returns:
            str: The generated code including increasing the offset of 'member_offset' by the real size of the array
        '''
        out = f'{self.indent}local members_size, members = parse_varint({tvb_name})\n'
        out += f'{self.indent}{member_offset} = {member_offset} + members_size\n'
        out += f'{self.indent}for i=1,members do\n'
        self.enter_indent()

        if m.type.subtype.identity == TypeIdentity.STRING:
            out += self.generate_string_parsing(m, message, tree_name, tvb_name, tvb_offset=member_offset)
            out += f'{self.indent}{member_offset} = {member_offset} + len_size + len\n'
        elif self.is_simple_integer(self.proto_field_type(m.type.subtype)) and m.type.subtype.identity != TypeIdentity.ARRAY:
            out += f'{self.indent}local member_size, member = parse_varint({tvb_name}({member_offset}))\n'

            type_prefix, type_suffix = self.wireshark_lua_type(m.type)
            out += f'{self.indent}local {m.name}_item = {tree_name}:add_le({self.proto_field_member_name(message, m)}, {tvb_name}({member_offset}, member_size), {type_prefix}member{type_suffix})\n'
            if m.type.identity in (TypeIdentity.ENUM, TypeIdentity.CONSTS) and "format" in m.type.attributes:
                format_attr = m.type.attributes['format'].value
                if format_attr == "bitfield":
                    for em in self.enums[m.type.name].members:
                        out += f"{self.indent}{m.name}_item:add_le({self.proto_field_member_name(message, m)}_{em.name}, {tvb_name}({member_offset}, member_size), {type_prefix}member{type_suffix})\n"
                        message.fields.append(f"{m.name}_{em.name}")

            out += f'{self.indent}{member_offset} = {member_offset} + member_size\n'
        elif m.type.subtype.identity == TypeIdentity.STRUCT:
            struct_def = self.structs[m.type.subtype.name]
            out += f'{self.indent}local struct_size = {self.proto_name(struct_def)}_struct_size({tvb_name}({member_offset}))\n'
            out += f'{self.indent}{self.proto_name(struct_def)}.dissector({tvb_name}({member_offset}, struct_size):tvb(), pinfo, {tree_name})\n'
            out += f'{self.indent}{member_offset} = {member_offset} + struct_size\n'
        else:
            print(f"unhandled {loc} member {m.type.subtype}[] {m.name} ({m.type})")
        self.leave_indent()
        out += f'{self.indent}end\n'

        return out

    def generate(self):
        '''
        Main code generation function
        '''

        out = '-- This file has been autogenerated, changes *will* be lost eventually...\n'
        out += self.generate_header()

        for unit in self.units:
            self.enums = {}
            self.messages = {}
            self.structs = {}

            for thing in unit.tokens:
                if type(thing) == Enum:
                    self.enums[thing.name] = thing
                    out += self.generate_consts(thing)
                elif type(thing) == Message:
                    out += self.generate_message(thing)
                elif type(thing) == NamespaceTag:
                    self.protohash = hashlib.shake_128(thing.name.encode('ascii')).hexdigest(4)
                    self.shortprotoname = thing.name.rsplit('::', 1)[-1]
                    self.protoname = thing.name
                    self.protocols[self.protohash] = self.protoname
                elif type(thing) == Struct:
                    out += self.generate_struct(thing)
                else:
                    print(f"unhandled type {str(type(thing))}")

            out += self.generate_protocol_metadata()

        out += self.generate_global_protocol_metadata()
        out += self.generate_footer()

        return out

    def generate_consts(self, enum):
        '''
        Generate enum value -> name lookup table
        '''

        i = 0
        out = f'local dis_{self.protohash}_{enum.name} = {{\n'
        self.enter_indent()

        for m in enum.members:
            if m.value is not None:
                i = m.value

            out += f'{self.indent}[{i}] = "{m.name}",\n'

            i += 1

        self.leave_indent()
        out += '}\n\n'

        return out

    def is_simple_integer(self, t):
        return t in ['byte', 'char', 'int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32', 'int64', 'uint64']

    def is_unsigned_integer(self, t):
        return t in ['byte', 'uint8', 'uint16', 'uint32', 'uint64']

    def bits_in_integer(self, t):
        if t.name in ['char', 'int8', 'uint8']:
            return 8
        elif t.name in ['int16', 'uint16']:
            return 16
        elif t.name in ['int32', 'uint32']:
            return 32
        elif t.name in ['int64', 'uint64']:
            return 64

    def is_dyn_pointer(self, m):
        return type(m) is TagsBlock or m.type.dynamic

    def count_dynamic(self, members):
        i = 0

        for m in members:
            if self.is_dyn_pointer(m): i += 1

        return i

    def wireshark_lua_type(self, t):
        '''
        Generate wireshark conversion for a given type
        '''

        if t.identity == TypeIdentity.INTEGER:
            if str(t) in ['byte', 'char', 'int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32']:
                return '', ''
            elif str(t) == 'uint64':
                return 'UInt64(', ')'
            elif str(t) == 'int64':
                return 'Int64(', ')'
        elif t.identity in (TypeIdentity.ENUM, TypeIdentity.CONSTS):
            return self.wireshark_lua_type(self.enums[t.name].type.subtype)
        elif t.identity == TypeIdentity.ARRAY:
            return self.wireshark_lua_type(t.subtype)

        raise Exception(f'Unhandled type {t.identity} for {t.name}')

    def proto_field_member_name(self, message: Message | Struct, member):
        '''
        Generates the name of the ProtoField object for a given message member
        '''
        return self.proto_field_name(message, member.name)

    def proto_field_name(self, message: Message | Struct, name: str):
        '''
        Generates the name of the ProtoField object for a given message and field name
        '''
        return f'{self.proto_name(message)}_{name}'

    def proto_name(self, message: Message | Struct):
        '''
        Genereates the name of the Proto object for a given message
        '''
        return f'dis_{self.protohash}_{message.name}'

    def proto_field_type(self, t):
        '''
        Generates the ProtoField type for a given type
        '''

        if t.identity == TypeIdentity.INTEGER:
            if t.name == "byte":
                return 'uint8'
            return t.name
        elif t.identity in (TypeIdentity.ENUM, TypeIdentity.CONSTS):
            return self.proto_field_type(self.enums[t.name].type.subtype)
        elif t.identity == TypeIdentity.STRING:
            return 'string'
        elif t.identity == TypeIdentity.ARRAY:
            if t.subtype.identity == TypeIdentity.INTEGER and t.subtype.name in ('uint8', 'byte', 'char'):
                return 'bytes'
            else:
                return self.proto_field_type(t.subtype)
        elif t.identity == TypeIdentity.STRUCT:
            return 'none'

        return f'UNKNOWN {t.identity}'

    def generate_protofield(self, message: Message | Struct, m: MessageMember) -> str:
        '''
        Generates the ProtoField definition for a given message member

        Parameters:
            message (Message): The message the member belongs to
            m (MessageMember): The member to generate the ProtoField definition for

        Returns:
            str: The generated ProtoField definition
        '''
        field_type = self.proto_field_type(m.type)
        out = ""

        extra_arg = ''

        if self.is_simple_integer(field_type) and m.type.identity != TypeIdentity.ARRAY:
            int_format = "base.DEC"

            if field_type == 'char':
                int_format = "base.HEX"

            if m.format and m.format.value == "hex" and self.is_unsigned_integer(field_type):
                int_format = "base.HEX"
            elif m.format and m.format.value == "octal":
                int_format = "base.OCT"
            elif m.format and m.format.value == "ipv4":
                return f'{self.proto_field_member_name(message, m)} = ProtoField.ipv4("{self.shortprotoname}_{message.name}.{m.name}", "{m.name}")\n'

            if m.type.identity in (TypeIdentity.ENUM, TypeIdentity.CONSTS):
                if "format" in m.type.attributes and m.type.attributes["format"].value == "bitfield":
                    for em in self.enums[m.type.name].members:
                        out += f'{self.indent}{self.proto_field_member_name(message, m)}_{em.name} = ProtoField.new("{em.name}", "{em.name}", ftypes.BOOLEAN, nil, {self.bits_in_integer(m.type.subtype)}, {em.value})\n'
                else:
                    extra_arg += f', dis_{self.protohash}_{m.type.name}'

            return f'{out}{self.proto_field_member_name(message, m)} = ProtoField.{field_type}("{self.shortprotoname}_{message.name}.{m.name}", "{m.name}", {int_format}{extra_arg})\n'
        else:
            return f'{out}{self.proto_field_member_name(message, m)} = ProtoField.{field_type}("{self.shortprotoname}_{message.name}.{m.name}", "{m.name}"{extra_arg})\n'

    def generate_message(self, message):
        message.fields = []
        tag_members = {}

        out = f'{self.proto_name(message)} = Proto("{self.shortprotoname}_{message.name}", "{self.protoname} {message.name} request")\n\n'

        head_members = flatten((m.members if type(m) is TagsBlock else [m] for m in message.head.members) if message.head is not None else [])
        tail_members = flatten((m.members if type(m) is TagsBlock else [m] for m in message.tail.members) if message.tail is not None else [])
        all_members = flatten([ head_members, tail_members ])

        for m in all_members:
            if m.tag:
                tag_members[m.tag] = m

            message.fields.append(m.name)

            out += self.generate_protofield(message, m)

        out += f'\n{self.indent}{self.proto_name(message)}_tags = {{\n'
        self.enter_indent()
        for tag, m in tag_members.items():
            out += f'{self.indent}[{tag.value}] = {self.proto_field_member_name(message, m)},\n'

        self.leave_indent()
        out += f'{self.indent}}}\n\n'

        out += f'{self.indent}function {self.proto_name(message)}.dissector(tvb, pinfo, tree)\n'
        self.enter_indent()

        member_offset = 0
        for m in head_members:
            if m.tag:
                continue

            out += f'{self.indent}local {m.name}_item = tree:add_le({self.proto_field_member_name(message, m)}, tvb({member_offset}, {m.type.fixed_size}))\n'
            if m.type.identity in (TypeIdentity.ENUM, TypeIdentity.CONSTS) and "format" in m.type.attributes:
                format_attr = m.type.attributes['format'].value
                if format_attr == "bitfield":
                    for em in self.enums[m.type.name].members:
                        out += f"{self.indent}{m.name}_item:add_le({self.proto_field_member_name(message, m)}_{em.name}, tvb:range({member_offset}, {m.type.fixed_size}))\n"
                        message.fields.append(f"{m.name}_{em.name}")

            member_offset += m.type.fixed_size

        out += self.generate_message_tag_members(message, tag_members)
        out += self.generate_message_tail_members(message, tail_members)

        self.leave_indent()
        out += 'end\n\n'

        fields_str = ", ".join([self.proto_field_name(message, f) for f in message.fields])
        out += f'{self.proto_name(message)}.fields = {{{fields_str}}}\n\n'

        self.messages[message.id] = message

        return out

    def generate_message_tag_members(self, message, tag_members):
        if not tag_members:
            return ''

        first_tag = True
        tags_offset = self.calculate_fixed_part_size('head', message.head.members, message) - 8
        out = f'{self.indent}local tags = tvb({tags_offset}, tvb:len() - {tags_offset}):tvb()\n'
        out += f'{self.indent}local size, tag = parse_varint(tags)\n'
        out += f'{self.indent}tags = tags(size, tags:len() - size):tvb()\n'
        out += f'{self.indent}while tag ~= 0 do\n'
        self.enter_indent()
        out += f'{self.indent}local value_size = 0\n'
        for tag, m in tag_members.items():
            if first_tag:
                first_tag = False
                out += f'{self.indent}if tag == {tag.value} then\n'
            else:
                out += f'{self.indent}elseif tag == {tag.value} then\n'
            self.enter_indent()

            if self.is_simple_integer(self.proto_field_type(m.type)) and m.type.identity != TypeIdentity.ARRAY:
                out += self.generate_integer_parsing(m, message, tvb_name='tags')
                out += f'{self.indent}value_size = varint_size\n'
            elif m.type.identity == TypeIdentity.STRING:
                out += self.generate_string_parsing(m, message, tvb_name='tags')
                out += f'{self.indent}value_size = len_size + len\n'
            elif m.type.identity == TypeIdentity.ARRAY:
                out += self.generate_array_parsing(m, message, 'tag', tvb_name='tags')
            else:
                print(f'unhandled type {m.type}')

            self.leave_indent()
        out += f'{self.indent}end\n'

        out += f'{self.indent}size, tag = parse_varint(tags(value_size))\n'
        out += f'{self.indent}tags = tags(value_size + size, tags:len() - size - value_size):tvb()\n'

        self.leave_indent()
        out += f'{self.indent}end\n'

        return out

    def generate_message_tail_members(self, message, tail_members):
        if not tail_members:
            return ''

        out = ''
        head_size = self.calculate_fixed_part_size('head', message.head.members, message) - 8
        tail_member_num = 0
        for m in tail_members:
            out += f'{self.indent}local dynoff{tail_member_num} = tvb({head_size + (tail_member_num * 8)}, 4):le_uint()\n'
            out += f'{self.indent}local tail = tvb({head_size} + dynoff{tail_member_num}, tvb:len() - ({head_size} + dynoff{tail_member_num}))\n'
            out += f'{self.indent}local value_size = 0\n'
            if m.type.identity == TypeIdentity.STRING:
                out += f'{self.indent}local len_size, len = parse_varint(tail)\n'
                out += f'{self.indent}if len_size < tail:len() then\n'
                self.enter_indent()
                out += f'{self.indent}tree:add({self.proto_field_member_name(message, m)}, tail(len_size, len), tail(len_size, len):string())\n'
                self.leave_indent()
                out += f'{self.indent}else\n'
                self.enter_indent()
                out += f'{self.indent}tree:add({self.proto_field_member_name(message, m)}, tail(0, len_size + len), "<empty>")\n'
                self.leave_indent()
                out += f'{self.indent}end\n'
            elif m.type.identity == TypeIdentity.ARRAY:
                out += self.generate_array_parsing(m, message, 'tail', tvb_name='tail')

                out += f'{self.indent}len = value_size\n'
            elif self.is_simple_integer(self.proto_field_type(m.type)):
                out += f'{self.indent}if len_size < tail:len() then\n'
                self.enter_indent()
                out += f'{self.indent}tree:add({self.proto_field_member_name(message, m)}, tail(len_size, len), parse_varint(tail(len_size, len)))\n'
                self.leave_indent()
                out += f'{self.indent}else\n'
                self.enter_indent()
                out += f'{self.indent}tree:add({self.proto_field_member_name(message, m)}, tail(0, len_size + len), "<empty>")\n'
                self.leave_indent()
                out += f'{self.indent}end\n'
            else:
                print(f"unhandled tail member {m.type} {m.name}")
            tail_member_num += 1

        return out

    def generate_struct(self, struct: Struct):
        '''
        Generates the Proto and dissector for a given struct
        '''

        self.structs[struct.name] = struct
        out = f'{self.indent}{self.proto_name(struct)} = Proto("{self.shortprotoname}_{struct.name}", "{struct.name}")\n'

        members = []
        for item in struct.members:
            out += self.generate_protofield(struct, item)
            members.append(item.name)

        fields_str = ", ".join([self.proto_field_name(struct, f) for f in members])
        out += f'{self.indent}{self.proto_name(struct)}.fields = {{{fields_str}}}\n'
        out += f'{self.indent}function {self.proto_name(struct)}.dissector(tvb, pinfo, tree)\n'
        self.enter_indent()
        out += f'{self.indent}local subtree = tree:add({self.proto_name(struct)}, tvb(0))\n'
        out += f'{self.indent}local offset = 0\n'
        for m in struct.members:
            if m.type.identity == TypeIdentity.STRING:
                out += self.generate_string_parsing(m, struct, tree_name='subtree', tvb_offset='offset')
                out += f'{self.indent}offset = offset + len_size + len\n'
            elif self.is_simple_integer(self.proto_field_type(m.type)) and m.type.identity != TypeIdentity.ARRAY:
                out += self.generate_integer_parsing(m, struct, tree_name='subtree')
                out += f'{self.indent}offset = offset + varint_size\n'
            else:
                print(f'unhandled struct member {m.type} {m.name}')
        self.leave_indent()
        out += 'end\n\n'

        # generate a function to calculate the struct's size
        out += f'{self.indent}function {self.proto_name(struct)}_struct_size(tvb)\n'
        self.enter_indent()
        out += f'{self.indent}local offset = 0\n'
        for m in struct.members:
            if m.type.identity == TypeIdentity.STRING:
                out += f'{self.indent}local len_size, len = parse_varint(tvb(offset))\n'
                out += f'{self.indent}offset = offset + len_size + len\n'
            elif self.is_simple_integer(self.proto_field_type(m.type)) and m.type.identity != TypeIdentity.ARRAY:
                if self.proto_field_type(m.type) in ('uint8', 'char'):
                    out += f'{self.indent}offset = offset + 1\n'
                else:
                    out += f'{self.indent}local varint_size, value = parse_varint(tvb(offset))\n'
                    out += f'{self.indent}offset = offset + varint_size\n'
            else:
                print(f'unhandled struct member {m.type} {m.name}')
        out += f'{self.indent}return offset\n'
        self.leave_indent()
        out += 'end\n\n'

        return out

    def generate_protocol_metadata(self):
        '''
        Generates the protocol metadata for the current bragi protocol dissector
        '''

        out = f'local dis_{self.protohash}_msgs = {{\n'
        self.enter_indent()

        for m in self.messages:
            out += f'{self.indent}[{m}] = {self.proto_name(self.messages[m])},\n'

        self.leave_indent()
        out += '}\n\n'

        out += f'local dis_{self.protohash}_message_ids = {{\n'
        self.enter_indent()

        for m in self.messages:
            out += f'{self.indent}[{m}] = "{self.messages[m].name}",\n'

        self.leave_indent()
        out += '}\n\n'

        return out

    def generate_global_protocol_metadata(self):
        '''
        Generates the global protocol metadata for the generated bragi protocol dissectors
        '''

        out = f'local protocol_mappings = {{\n'
        self.enter_indent()
        for p in self.protocols:
            out += f'{self.indent}[0x{p}] = "{self.protocols[p]}",\n'
        self.leave_indent()
        out += '}\n\n'

        out += f'local protocol_message_ids = {{\n'
        self.enter_indent()
        for p in self.protocols:
            out += f'{self.indent}[0x{p}] = dis_{p}_message_ids,\n'
        self.leave_indent()
        out += '}\n\n'

        out += f'local protocols = {{\n'
        self.enter_indent()
        for p in self.protocols:
            out += f'{self.indent}[0x{p}] = dis_{p}_msgs,\n'
        self.leave_indent()
        out += '}\n\n'

        return out

    def determine_pointer_size(self, what, size):
        if what != 'head':
            return 8

        if size < 256:
            return 1
        elif size < 65536:
            return 2
        elif size < 4294967296:
            return 4
        elif size < 18446744073709551616:
            return 8
        else:
            return None

    def calculate_fixed_part_size(self, what, members, parent):
        i = 8 if what == 'head' else 0

        ptr_size = self.determine_pointer_size(what, parent.head.size if what == 'head' else None)

        for m in members:
            if not self.is_dyn_pointer(m):
                assert m.type.fixed_size
                i += m.type.fixed_size
            else:
                i += ptr_size if ptr_size else 0

        return i



#!/usr/bin/env python3

import argparse

from bragi.parser import CompilationUnit
from bragi.cpp_generator import CodeGenerator as CppCodeGenerator
from bragi.wireshark_generator import CodeGenerator as WiresharkCodeGenerator
from bragi.rust_generator import CodeGenerator as RustCodeGenerator

parser = argparse.ArgumentParser(prog = 'bragi', description = 'Bragi IDL to C++ compiler')
parser.add_argument('input', nargs='+', help='input file', type=argparse.FileType('r'))
parser.add_argument('-o', '--output', help='output file', type=str)
subparsers = parser.add_subparsers(required=True, dest='language')

cpp_parser = subparsers.add_parser('cpp')
cpp_parser.add_argument('-l', '--lib', nargs=1, help='C++ library to use', choices=['frigg', 'stdc++'], default='libc++')
cpp_parser.add_argument('--protobuf', help='Generate protobuf compatibilty methods (SerializeAsString/ParseFromArray)', action='store_true')

ws_parser = subparsers.add_parser('wireshark')

rust_parser = subparsers.add_parser('rust')

def main():
	args = parser.parse_args()

	inputs = []
	output = args.output

	for source in args.input:
		code = source.read()
		unit = CompilationUnit(source.name, code)
		unit.process()
		unit.verify()
		inputs.append(unit)

	if args.language == 'cpp':
		lib = args.lib[0]
		generator = CppCodeGenerator(inputs, lib, protobuf_compat = args.protobuf)
	elif args.language == 'wireshark':
		generator = WiresharkCodeGenerator(inputs)
	elif args.language == 'rust':
		generator = RustCodeGenerator(inputs)

	with open(output, "w") as o:
		o.write(generator.generate())



"""This module implements a LALR(1) Parser
"""
# Author: Erez Shinan (2017)
# Email : erezshin@gmail.com
from typing import Dict, Any, Optional
from ..lexer import Token, LexerThread
from ..utils import Serialize
from ..common import ParserConf, ParserCallbacks

from .lalr_analysis import LALR_Analyzer, IntParseTable, ParseTableBase
from .lalr_interactive_parser import InteractiveParser
from lark.exceptions import UnexpectedCharacters, UnexpectedInput, UnexpectedToken
from .lalr_parser_state import ParserState, ParseConf

###{standalone

class LALR_Parser(Serialize):
    def __init__(self, parser_conf: ParserConf, debug: bool=False, strict: bool=False):
        analysis = LALR_Analyzer(parser_conf, debug=debug, strict=strict)
        analysis.compute_lalr()
        callbacks = parser_conf.callbacks

        self._parse_table = analysis.parse_table
        self.parser_conf = parser_conf
        self.parser = _Parser(analysis.parse_table, callbacks, debug)

    @classmethod
    def deserialize(cls, data, memo, callbacks, debug=False):
        inst = cls.__new__(cls)
        inst._parse_table = IntParseTable.deserialize(data, memo)
        inst.parser = _Parser(inst._parse_table, callbacks, debug)
        return inst

    def serialize(self, memo: Any = None) -> Dict[str, Any]:
        return self._parse_table.serialize(memo)

    def parse_interactive(self, lexer: LexerThread, start: str):
        return self.parser.parse(lexer, start, start_interactive=True)

    def parse(self, lexer, start, on_error=None):
        try:
            return self.parser.parse(lexer, start)
        except UnexpectedInput as e:
            if on_error is None:
                raise

            while True:
                if isinstance(e, UnexpectedCharacters):
                    s = e.interactive_parser.lexer_thread.state
                    p = s.line_ctr.char_pos

                if not on_error(e):
                    raise e

                if isinstance(e, UnexpectedCharacters):
                    # If user didn't change the character position, then we should
                    if p == s.line_ctr.char_pos:
                        s.line_ctr.feed(s.text.text[p:p+1])

                try:
                    return e.interactive_parser.resume_parse()
                except UnexpectedToken as e2:
                    if (isinstance(e, UnexpectedToken)
                        and e.token.type == e2.token.type == '$END'
                        and e.interactive_parser == e2.interactive_parser):
                        # Prevent infinite loop
                        raise e2
                    e = e2
                except UnexpectedCharacters as e2:
                    e = e2


class _Parser:
    parse_table: ParseTableBase
    callbacks: ParserCallbacks
    debug: bool

    def __init__(self, parse_table: ParseTableBase, callbacks: ParserCallbacks, debug: bool=False):
        self.parse_table = parse_table
        self.callbacks = callbacks
        self.debug = debug

    def parse(self, lexer: LexerThread, start: str, value_stack=None, state_stack=None, start_interactive=False):
        parse_conf = ParseConf(self.parse_table, self.callbacks, start)
        parser_state = ParserState(parse_conf, lexer, state_stack, value_stack)
        if start_interactive:
            return InteractiveParser(self, parser_state, parser_state.lexer)
        return self.parse_from_state(parser_state)


    def parse_from_state(self, state: ParserState, last_token: Optional[Token]=None):
        """Run the main LALR parser loop

        Parameters:
            state - the initial state. Changed in-place.
            last_token - Used only for line information in case of an empty lexer.
        """
        try:
            token = last_token
            for token in state.lexer.lex(state):
                assert token is not None
                state.feed_token(token)

            end_token = Token.new_borrow_pos('$END', '', token) if token else Token('$END', '', 0, 1, 1)
            return state.feed_token(end_token, True)
        except UnexpectedInput as e:
            try:
                e.interactive_parser = InteractiveParser(self, state, state.lexer)
            except NameError:
                pass
            raise e
        except Exception as e:
            if self.debug:
                print("")
                print("STATE STACK DUMP")
                print("----------------")
                for i, s in enumerate(state.state_stack):
                    print('%d)' % i , s)
                print("")

            raise
###}



from .tokens import *
from .types import *

import hashlib

class StdlibTraits:
    def needs_allocator(self):
        return False
    def allocator_argument(self):
        return ''
    def allocator_parameter(self):
        return ''
    def array(self):
        return 'std::array'
    def vector(self):
        return 'std::vector'
    def string(self):
        return 'std::string'
    def assert_func(self):
        return 'assert'
    def includes(self):
        return ['<stdint.h>', '<stddef.h>', '<array>', '<vector>', '<cassert>', '<optional>', '<string>']

class FriggTraits:
    def needs_allocator(self):
        return True
    def allocator_argument(self):
        return 'Allocator allocator = Allocator()'
    def allocator_parameter(self):
        return 'allocator'
    def array(self):
        return 'std::array'
    def vector(self):
        return 'frg::vector'
    def string(self):
        return 'frg::string'
    def assert_func(self):
        return 'FRG_ASSERT'
    def includes(self):
        return ['<stdint.h>', '<stddef.h>', '<array>', '<frg/vector.hpp>', '<frg/macros.hpp>', '<frg/optional.hpp>', '<frg/string.hpp>']

flatten = lambda l: [item for sublist in l for item in sublist]

class CodeGenerator:
    def __init__(self, unit, stdlib, protobuf_compat = False):
        self.units = unit
        self.protobuf_compat = protobuf_compat
        self.stdlib_traits = None
        self.indent_depth = 0
        self.indent = ''

        if stdlib == 'stdc++':
            self.stdlib_traits = StdlibTraits()
        elif stdlib == 'frigg':
            self.stdlib_traits = FriggTraits()
        else:
            raise AttributeError('invalid standard library')

        self.current_ns = None

    def enter_indent(self):
        self.indent_depth += 1
        self.indent = '\t' * self.indent_depth
        return self.indent

    def leave_indent(self):
        self.indent_depth -= 1
        self.indent = '\t' * self.indent_depth
        return self.indent

    def generate(self):
        out = '// This file has been autogenerated, changes *will* be lost eventually...\n'
        out += '#pragma once\n'

        for i in self.stdlib_traits.includes():
            out += f'#include {i}\n'

        out += '#include <bragi/internals.hpp>\n\n'

        for unit in self.units:
            for thing in unit.tokens:
                if type(thing) == NamespaceTag:
                    out += self.switch_ns(thing)
                    protohash = hashlib.shake_128(thing.name.encode('ascii')).hexdigest(4)
                    out += f'{self.indent}const char protocol_hash[] = "0x{protohash}";\n\n'
                if type(thing) == UsingTag:
                    out += self.generate_using(thing)
                if type(thing) == Enum and thing.mode == "enum":
                    out += self.generate_enum(thing)
                if type(thing) == Enum and thing.mode == "consts":
                    out += self.generate_consts(thing)
                if type(thing) == Message:
                    out += self.generate_message(thing)
                if type(thing) == Struct:
                    out += self.generate_struct(thing)
                if type(thing) == Group:
                    for m in thing.members:
                        out += self.generate_message(m)

        out += self.finalize_ns()

        return out

    def switch_ns(self, ns):
        out = ''

        if self.current_ns:
            out += self.finalize_ns()

        out += f'namespace {ns.name} {{\n\n'

        self.current_ns = ns
        return out

    def finalize_ns(self):
        if self.current_ns:
            return f'}} // namespace {self.current_ns.name}\n\n'

        return ''

    def generate_consts(self, enum):
        out = f'{self.indent}namespace {enum.name} {{\n'
        i = 0
        self.enter_indent()

        for m in enum.members:
            if m.value is not None:
                i = m.value

            out += f'{self.indent}inline constexpr {self.generate_type(enum.type)} {m.name} = {i};\n'

            i += 1

        self.leave_indent()
        return out + f'{self.indent}}} // namespace {enum.name}\n\n'

    def generate_enum(self, enum):
        out = f'{self.indent}enum class {enum.name} : int32_t {{\n'
        i = 0
        self.enter_indent()

        for m in enum.members:
            if m.value is not None:
                i = m.value

            out += f'{self.indent}{m.name} = {i},\n'

            i += 1

        self.leave_indent()
        return out + f'{self.indent}}}; // enum class {enum.name}\n\n'

    def make_ns_name(self, full):
        a, b, c = full.rpartition('::')

        if b == '':
            return (self.current_ns.name if self.current_ns else '', None, full)

        return (None, a, c)

    def generate_using(self, using):
        from_a, from_b, from_name = self.make_ns_name(using.from_name)
        to_a, to_b, to_name = self.make_ns_name(using.to_name)
        from_ns = from_a if from_a else from_b

        from_full = f'::{from_ns}::{from_name}'

        if not to_a:
            last_ns = self.current_ns
            out = self.finalize_ns()
            self.current_ns = None

            if to_b != '':
                out += f'namespace {to_b} {{\n\n'

            if self.stdlib_traits.needs_allocator():
                out += f'{self.indent}template <typename Allocator>\n'
                out += f'{self.indent}using {to_name} = {from_full}<Allocator>;\n\n'
            else:
                out += f'{self.indent}using {to_name} = {from_full};\n\n'

            if to_b != '':
                out += f'}} // namespace {to_b}\n\n'
            out += self.switch_ns(last_ns)
            return out
        else:
            out = ''

            if self.stdlib_traits.needs_allocator():
                out += f'{self.indent}template <typename Allocator>\n'
                out += f'{self.indent}using {to_name} = {from_full}<Allocator>;\n\n'
            else:
                out += f'{self.indent}using {to_name} = {from_full};\n\n'

            return out

    def check_needs_allocator(self, t):
        if t.identity is TypeIdentity.ARRAY and t.n_elements is not None:
            return False

        return (t.identity in [TypeIdentity.STRING, TypeIdentity.ARRAY, TypeIdentity.STRUCT]
                or t.dynamic) and self.stdlib_traits.needs_allocator()

    def generate_type(self, t):
        if t.identity is TypeIdentity.INTEGER:
            if t.name == 'char':
                return 'char'
            return f'{"u" if not t.signed else ""}int{t.fixed_size * 8}_t'
        elif t.identity is TypeIdentity.CONSTS:
            return self.generate_type(t.subtype)
        elif t.identity is TypeIdentity.ENUM:
            return t.name
        elif t.identity is TypeIdentity.STRING:
            return f'{self.stdlib_traits.string()}{"<Allocator>" if self.stdlib_traits.needs_allocator() else ""}'
        elif t.identity is TypeIdentity.ARRAY:
            if t.n_elements is None:
                return f'{self.stdlib_traits.vector()}<{self.generate_type(t.subtype)}{", Allocator" if self.stdlib_traits.needs_allocator() else ""}>'
            else:
                return f'{self.stdlib_traits.array()}<{self.generate_type(t.subtype)}, {t.n_elements}>'
        elif t.identity is TypeIdentity.STRUCT:
            return f'{t.name}{"<Allocator>" if self.stdlib_traits.needs_allocator() else ""}'
        else:
            raise RuntimeError('unknown type in generate_type')

    def is_simple_integer(self, t):
        return t in ['char', 'int8_t', 'uint8_t', 'int16_t', 'uint16_t', 'int32_t', 'uint32_t', 'int64_t', 'uint64_t']

    def is_dyn_pointer(self, m):
        return type(m) is TagsBlock or m.type.dynamic

    def count_dynamic(self, members):
        i = 0

        for m in members:
            if self.is_dyn_pointer(m): i += 1

        return i

    def emit_stmt_checked(self, stmt):
        out = f'{self.indent}if (!{stmt})\n'

        self.enter_indent()
        out += f'{self.indent}return false;\n'
        self.leave_indent()

        return out

    def emit_assert_that(self, stmt):
        return f'{self.indent}{self.stdlib_traits.assert_func()}({stmt});\n'

    def determine_pointer_size(self, what, size):
        if what != 'head':
            return 8

        if size < 256:
            return 1
        elif size < 65536:
            return 2
        elif size < 4294967296:
            return 4
        elif size < 18446744073709551616:
            return 8
        else:
            return None

    def calculate_fixed_part_size(self, what, members, parent):
        i = 8 if what == 'head' else 0

        ptr_size = self.determine_pointer_size(what, parent.head.size if what == 'head' else None)

        for m in members:
            if not self.is_dyn_pointer(m):
                assert m.type.fixed_size
                i += m.type.fixed_size
            else:
                i += ptr_size

        return i

    def emit_calculate_dynamic_size_of_member(self, into, member):
        if type(member) is TagsBlock:
            out = ''
            for child in member.members:
                out += f'{self.indent}if (p_{child.name}) {{\n'
                self.enter_indent()
                out += f'{self.indent}{into} += bragi::detail::size_of_varint({child.tag.value});\n'
                out += self.emit_calculate_dynamic_size_of_member(into, child)
                self.leave_indent()
                out += f'{self.indent}}}\n\n'
            out += f'{self.indent}{into} += bragi::detail::size_of_varint(0);\n'
            return out
        else:
            return self.emit_calculate_dynamic_size_of_member_internal(into, f'm_{member.name}', member.type, 0)

    def emit_calculate_dynamic_size_of_member_internal(self, into, expr, expr_type, array_depth):
        if expr_type.identity in {TypeIdentity.INTEGER, TypeIdentity.CONSTS}:
            if expr_type.fixed_size == 1:
                return f'{self.indent}{into} += 1;\n'
            return f'{self.indent}{into} += bragi::detail::size_of_varint({expr});\n'
        elif expr_type.identity is TypeIdentity.ENUM:
            return f'{self.indent}{into} += bragi::detail::size_of_varint(static_cast<int32_t>({expr}));\n'
        elif expr_type.identity is TypeIdentity.STRING:
            out = f'{self.indent}{into} += bragi::detail::size_of_varint({expr}.size());\n'
            return out + f'{self.indent}{into} += {expr}.size();\n'
        elif expr_type.identity is TypeIdentity.ARRAY:
            out = f'{self.indent}{into} += bragi::detail::size_of_varint({expr}.size());\n'
            out += f'{self.indent}for (size_t i{array_depth} = 0; i{array_depth} < {expr}.size(); i{array_depth}++) {{\n'
            self.enter_indent()
            out += self.emit_calculate_dynamic_size_of_member_internal(into, f'{expr}[i{array_depth}]', expr_type.subtype, array_depth + 1)
            self.leave_indent()
            return out + f'{self.indent}}}\n'
        elif expr_type.identity is TypeIdentity.STRUCT:
            return f'{self.indent}{into} += {expr}.size_of_body();\n'
        else:
            raise RuntimeError('unexpected member type')

    def emit_determine_dyn_off_for(self, skip, prev, n):
        out = ''
        into = f'dyn_offs[{n}]'

        if n > 0:
            out += f'{self.indent}{into} = dyn_offs[{n - 1}];\n'
        else:
            out += f'{self.indent}{into} = {skip};\n'

        if not prev:
            return out + '\n'
        else:
            return out + self.emit_calculate_dynamic_size_of_member(into, prev) + '\n'

    def emit_calculate_size_of(self, what, members, parent):
        out = f'{self.indent}size_t size_of_{what}() {{\n'
        self.enter_indent()

        out += f'{self.indent}size_t size = {self.calculate_fixed_part_size(what, members, parent)};\n'

        for member in filter(self.is_dyn_pointer, members):
            out += self.emit_calculate_dynamic_size_of_member('size', member)

        out += f'\n{self.indent}return size;\n'

        self.leave_indent()
        out += f'{self.indent}}}\n'

        return out + '\n'

    def emit_struct_calculate_size_of(self, members, parent):
        out = f'{self.indent}size_t size_of_body() {{\n'
        self.enter_indent()

        out += f'{self.indent}size_t size = 0;\n'

        for member in members:
            out += self.emit_calculate_dynamic_size_of_member('size', member)

        out += f'\n{self.indent}return size;\n'

        self.leave_indent()
        out += f'{self.indent}}}\n'

        return out + '\n'

    def emit_stub_calculate_size_of(self, what):
        out = f'{self.indent}size_t size_of_{what}() {{\n'
        self.enter_indent()
        out += f'{self.indent}return {8 if what == "head" else 0};\n'
        self.leave_indent()

        out += f'{self.indent}}}\n'

        return out + '\n'

    def determine_pointer_type(self, what, size):
        size = self.determine_pointer_size(what, size)
        if not size:
            return None

        return f'uint{size * 8}_t'

    class FixedEncoder:
        def __init__(self, parent):
            self.parent = parent
            self.nth_dynamic = 0

        def emit_encode_in_fixed(self, member, ptr_type):
            is_tags = type(member) is TagsBlock
            return self.emit_encode_in_fixed_internal(f'm_{member.name if not is_tags else "tags"}',
                    member.type if not is_tags else None, is_tags, ptr_type, 0)

        def emit_encode_in_fixed_default(self, expr_type, array_depth):
            assert expr_type and not expr_type.dynamic
            if expr_type.identity in {TypeIdentity.INTEGER, TypeIdentity.CONSTS}:
                return self.parent.emit_stmt_checked(f'sr.write_integer<{self.parent.generate_type(expr_type)}>(wr, 0)')
            elif expr_type.identity is TypeIdentity.ENUM:
                return self.parent.emit_stmt_checked(f'sr.write_integer<int32_t>(wr, static_cast<int32_t>(0))')
            elif expr_type.identity is TypeIdentity.ARRAY:
                assert not expr_type.subtype.dynamic
                assert expr_type.n_elements

                out = f'{self.parent.indent}for (size_t i{array_depth} = 0; i{array_depth} < {expr_type.n_elements}; i{array_depth}++) {{\n'
                self.parent.enter_indent()
                out += self.emit_encode_in_fixed_default(expr_type.subtype, array_depth + 1)
                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'

                return out
            else:
                assert member.type.identity not in {TypeIdentity.STRING, TypeIdentity.STRUCT}

        def emit_encode_in_fixed_internal(self, expr, expr_type, is_tags, ptr_type, array_depth):
            if is_tags or expr_type.dynamic:
                out = self.parent.emit_stmt_checked(f'sr.write_integer<{ptr_type}>(wr, dyn_offs[{self.nth_dynamic}])')
                self.nth_dynamic += 1
                return out
            elif expr_type.identity in {TypeIdentity.INTEGER, TypeIdentity.CONSTS}:
                return self.parent.emit_stmt_checked(f'sr.write_integer<{self.parent.generate_type(expr_type)}>(wr, {expr})')
            elif expr_type.identity is TypeIdentity.ENUM:
                return self.parent.emit_stmt_checked(f'sr.write_integer<int32_t>(wr, static_cast<int32_t>({expr}))')
            elif expr_type.identity is TypeIdentity.ARRAY:
                assert not expr_type.subtype.dynamic
                assert expr_type.n_elements

                out = f'{self.parent.indent}for (size_t i{array_depth} = 0; i{array_depth} < {expr_type.n_elements}; i{array_depth}++) {{\n'
                self.parent.enter_indent()
                out += f'{self.parent.indent}if (i{array_depth} < {expr}.size()) {{\n'
                self.parent.enter_indent()
                out += self.emit_encode_in_fixed_internal(f'{expr}[i{array_depth}]', expr_type.subtype, False, None, array_depth + 1)
                self.parent.leave_indent()
                out += f'{self.parent.indent}}} else {{\n'
                self.parent.enter_indent()
                out += self.emit_encode_in_fixed_default(expr_type.subtype, array_depth + 1)
                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'
                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'

                return out
            else:
                raise RuntimeError('unexpected variable type')

    class DynamicEncoder:
        def __init__(self, parent):
            self.parent = parent

        def emit_encode_in_dynamic(self, member):
            if type(member) is TagsBlock:
                out = ''
                for child in member.members:
                    out += f'{self.parent.indent}if (p_{child.name}) {{\n'
                    self.parent.enter_indent()
                    out += self.parent.emit_stmt_checked(f'sr.write_varint(wr, {child.tag.value})')
                    out += self.emit_encode_in_dynamic(child)
                    self.parent.leave_indent()
                    out += f'{self.parent.indent}}}\n\n'
                out += self.parent.emit_stmt_checked(f'sr.write_varint(wr, 0)')
                return out
            else:
                return self.emit_encode_in_dynamic_internal(f'm_{member.name}', member.type, 0)

        def emit_encode_in_dynamic_internal(self, expr, expr_type, array_depth):
            if expr_type.identity in {TypeIdentity.INTEGER, TypeIdentity.CONSTS}:
                if expr_type.fixed_size == 1:
                    return self.parent.emit_stmt_checked(f'sr.write_integer<{self.parent.generate_type(expr_type)}>(wr, {expr})')
                return self.parent.emit_stmt_checked(f'sr.write_varint(wr, static_cast<{self.parent.generate_type(expr_type)}>({expr}))')
            elif expr_type.identity is TypeIdentity.ENUM:
                return self.parent.emit_stmt_checked(f'sr.write_varint(wr, static_cast<int32_t>({expr}))')
            elif expr_type.identity in {TypeIdentity.ARRAY, TypeIdentity.STRING}:
                out = self.parent.emit_stmt_checked(f'sr.write_varint(wr, {expr}.size())')
                out += f'{self.parent.indent}for (size_t i{array_depth} = 0; i{array_depth} < {expr}.size(); i{array_depth}++) {{\n'
                self.parent.enter_indent()
                out += self.emit_encode_in_dynamic_internal(f'{expr}[i{array_depth}]', expr_type.subtype, array_depth + 1)
                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'

                return out
            elif expr_type.identity is TypeIdentity.STRUCT:
                return self.parent.emit_stmt_checked(f'{expr}.encode_body(wr, sr)')
            else:
                raise RuntimeError('unexpected variable type')

    def emit_part_encoder(self, what, parent, members):
        out = f'{self.indent}template <typename Writer>\n'
        out += f'{self.indent}bool encode_{what}(Writer &wr) {{\n'
        self.enter_indent()

        out += f'{self.indent}(void)wr;\n'
        out += f'{self.indent}bragi::serializer sr; (void)sr;\n'

        fixed_size = self.calculate_fixed_part_size(what, members, parent) if members else None
        ptrs = [i for i in members if self.is_dyn_pointer(i)] if members else None
        ptr_type = self.determine_pointer_type(what, parent.head.size if what == 'head' else None) if parent else None

        if ptrs:
            if len(ptrs) > 0:
                out += f'{self.indent}{ptr_type} dyn_offs[{len(ptrs)}];\n'

        out += '\n'

        if ptrs:
            for i, m in enumerate(ptrs):
                out += self.emit_determine_dyn_off_for(fixed_size, ptrs[i - 1] if i > 0 else None, i)

        if what == 'head':
            out += f'{self.indent}// Encode ID\n'
            out += self.emit_stmt_checked(f'sr.template write_integer<uint32_t>(wr, message_id)')

            out += f'{self.indent}// Encode tail size\n'
            out += self.emit_stmt_checked(f'sr.template write_integer<uint32_t>(wr, size_of_tail())')

        if members:
            fixed_enc = self.FixedEncoder(self)
            dyn_enc = self.DynamicEncoder(self)
            for m in members:
                out += fixed_enc.emit_encode_in_fixed(m, ptr_type) + '\n'

            for m in ptrs:
                out += dyn_enc.emit_encode_in_dynamic(m)

        out += f'{self.indent}return true;\n'
        self.leave_indent()
        out += f'{self.indent}}}\n\n'

        return out

    def emit_struct_encoder(self, parent, members):
        out = f'{self.indent}template <typename Writer>\n'
        out += f'{self.indent}bool encode_body(Writer &wr, bragi::serializer &sr) {{\n'
        self.enter_indent()

        out += f'{self.indent}(void)wr;\n'
        out += f'{self.indent}(void)sr;\n'
        out += '\n'

        dyn_enc = self.DynamicEncoder(self)
        for m in members:
            out += dyn_enc.emit_encode_in_dynamic(m)

        out += f'{self.indent}return true;\n'
        self.leave_indent()
        out += f'{self.indent}}}\n\n'

        return out

    class Decoder:
        def __init__(self, parent):
            self.parent = parent

        def emit_decode_member(self, member, ptr_type):
            if type(member) is TagsBlock or member.type.dynamic:
                out = self.parent.emit_stmt_checked(f'de.read_integer<{ptr_type}>(rd, ptr)')
                out += f'{self.parent.indent}de.push_index(ptr);\n'
                out += self.emit_decode_dynamic(member)
                out += f'{self.parent.indent}de.pop_index();\n'
                return out
            else:
                out = self.emit_decode_fixed_internal(f'm_{member.name}', member.type, 0)
                out += f'{self.parent.indent}p_{member.name} = true;\n'
                return out

        def emit_decode_fixed_internal(self, expr, expr_type, array_depth):
            assert not expr_type.dynamic
            if expr_type.identity in {TypeIdentity.INTEGER, TypeIdentity.CONSTS}:
                out = self.parent.emit_stmt_checked(f'de.read_integer<{self.parent.generate_type(expr_type)}>(rd, {expr})')
                return out
            elif expr_type.identity is TypeIdentity.ENUM:
                out = f'{self.parent.indent}{{\n'
                self.parent.enter_indent()
                out += f'{self.parent.indent}int32_t tmp;\n'
                out += self.parent.emit_stmt_checked(f'de.read_integer<int32_t>(rd, tmp)')
                out += f'{self.parent.indent}{expr} = static_cast<{self.parent.generate_type(expr_type)}>(tmp);\n'
                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'
                return out
            elif expr_type.identity is TypeIdentity.ARRAY:
                assert not expr_type.subtype.dynamic
                assert expr_type.n_elements

                out = ''

                out += f'{self.parent.indent}for (size_t i{array_depth} = 0; i{array_depth} < {expr_type.n_elements}; i{array_depth}++) {{\n'
                self.parent.enter_indent()
                out += self.emit_decode_fixed_internal(f'{expr}[i{array_depth}]', expr_type.subtype, array_depth + 1)
                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'

                return out
            else:
                raise RuntimeError('unexpected variable type')

        def emit_decode_dynamic(self, member):
            if type(member) is TagsBlock:
                out = f'{self.parent.indent}{{\n'
                self.parent.enter_indent()
                out += f'{self.parent.indent}uint64_t tag;\n'

                out += f'{self.parent.indent}do {{\n'
                self.parent.enter_indent()
                out += self.parent.emit_stmt_checked(f'de.read_varint(rd, tag)')

                out += f'{self.parent.indent}switch (tag) {{\n'
                self.parent.enter_indent()

                out += f'{self.parent.indent}case 0: break;\n'

                for child in member.members:
                    out += f'{self.parent.indent}case {child.tag.value}:\n'
                    self.parent.enter_indent()
                    out += self.emit_decode_dynamic_internal(f'm_{child.name}', child.type, 0)
                    out += f'{self.parent.indent}p_{child.name} = true;\n'
                    out += f'{self.parent.indent}break;\n'
                    self.parent.leave_indent()

                out += f'{self.parent.indent}default:\n'
                self.parent.enter_indent()
                out += self.parent.emit_assert_that('!"Unknown tag!"')
                self.parent.leave_indent()

                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'
                self.parent.leave_indent()
                out += f'{self.parent.indent}}} while (tag);\n'

                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'
                return out
            else:
                out = self.emit_decode_dynamic_internal(f'm_{member.name}', member.type, 0)
                out += f'{self.parent.indent}p_{member.name} = true;\n'
                return out

        def emit_decode_dynamic_internal(self, expr, expr_type, array_depth):
            if expr_type.identity in {TypeIdentity.INTEGER, TypeIdentity.CONSTS, TypeIdentity.ENUM}:
                if expr_type.fixed_size == 1:
                    return self.parent.emit_stmt_checked(f'de.read_integer<{self.parent.generate_type(expr_type)}>(rd, {expr})')

                out = f'{self.parent.indent}{{\n'
                self.parent.enter_indent()
                out += f'{self.parent.indent}uint64_t tmp;\n'
                out += self.parent.emit_stmt_checked(f'de.read_varint(rd, tmp)')
                out += f'{self.parent.indent}{expr} = static_cast<{self.parent.generate_type(expr_type)}>(tmp);\n'
                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'
                return out
            elif expr_type.identity in {TypeIdentity.STRING, TypeIdentity.ARRAY}:
                out = f'{self.parent.indent}{{\n'
                self.parent.enter_indent()
                out += f'{self.parent.indent}uint64_t size;\n'
                out += self.parent.emit_stmt_checked(f'de.read_varint(rd, size)')

                target_size = 'size'

                if expr_type.identity is TypeIdentity.ARRAY and expr_type.n_elements:
                    target_size = expr_type.n_elements

                if not (expr_type.identity is TypeIdentity.ARRAY and expr_type.n_elements is not None):
                    if self.parent.check_needs_allocator(expr_type.subtype):
                        out += f'{self.parent.indent}{expr}.resize({target_size}, allocator);\n'
                    else:
                        out += f'{self.parent.indent}{expr}.resize({target_size});\n'
                out += f'{self.parent.indent}for (size_t i{array_depth} = 0; i{array_depth} < {target_size}; i{array_depth}++)\n'
                if target_size != 'size':
                    self.parent.enter_indent()
                    out += f'{self.parent.indent}if (i{array_depth} < size)\n'

                self.parent.enter_indent()
                out += self.emit_decode_dynamic_internal(f'{expr}[i{array_depth}]', expr_type.subtype, array_depth + 1)
                self.parent.leave_indent()

                if target_size != 'size':
                    self.parent.leave_indent()

                self.parent.leave_indent()
                out += f'{self.parent.indent}}}\n'

                return out
            elif expr_type.identity is TypeIdentity.STRUCT:
                return self.parent.emit_stmt_checked(f'{expr}.decode_body(rd, de)')
            else:
                raise RuntimeError('unexpected variable type')

    def emit_part_decoder(self, what, parent, members):
        out = f'{self.indent}template <typename Reader>\n'
        out += f'{self.indent}bool decode_{what}(Reader &rd) {{\n'
        self.enter_indent()

        out += f'{self.indent}(void)rd;\n'
        out += f'{self.indent}bragi::deserializer de; (void)de;\n'

        if members:
            ptr_type = self.determine_pointer_type(what, parent.head.size if what == 'head' else None)
            out += f'{self.indent}{ptr_type} ptr; (void)ptr;\n'

        if what == 'head':
            out += f'{self.indent}{{\n'
            self.enter_indent()
            out += f'{self.indent}uint32_t tmp;\n'
            out += f'{self.indent}// Decode and check ID\n'
            out += self.emit_stmt_checked(f'de.read_integer<uint32_t>(rd, tmp)')
            out += self.emit_stmt_checked('(tmp == message_id)')
            out += '\n'

            out += f'{self.indent}// Decode and ignore tail size\n'
            out += self.emit_stmt_checked(f'de.read_integer<uint32_t>(rd, tmp)')
            self.leave_indent()
            out += f'{self.indent}}}\n'
            out += '\n'

        dec = self.Decoder(self)

        if members:
            ptr_type = self.determine_pointer_type(what, parent.head.size if what == 'head' else None)

            for m in members:
                out += dec.emit_decode_member(m, ptr_type)
                out += '\n'

        out += f'{self.indent}return true;\n'
        self.leave_indent()
        out += f'{self.indent}}}\n\n'

        return out

    def emit_struct_decoder(self, parent, members):
        out = f'{self.indent}template <typename Reader>\n'
        out += f'{self.indent}bool decode_body(Reader &rd, bragi::deserializer &de) {{\n'
        self.enter_indent()

        out += f'{self.indent}(void)rd;\n'
        out += f'{self.indent}(void)de;'
        out += '\n'

        dec = self.Decoder(self)

        for m in members:
            out += dec.emit_decode_dynamic(m)
            out += '\n'

        out += f'{self.indent}return true;\n'
        self.leave_indent()
        out += f'{self.indent}}}\n\n'

        return out

    # Protobuf compatibilty code
    def emit_serialize_as_string(self):
        out = ''

        if type(self.stdlib_traits) is FriggTraits:
            out = f'{self.indent}void SerializeToString(frg::string<Allocator> *str) {{\n'
            self.enter_indent()

            out += f'{self.indent}str->resize(size_of_head());\n'
            out += f'{self.indent}bragi::limited_writer wr{{str->data(), str->size()}};\n\n'
            out += self.emit_assert_that('encode_head(wr)')

            self.leave_indent()
            out += f'{self.indent}}}\n\n'
        else:
            out = f'{self.indent}std::string SerializeAsString() {{\n'
            self.enter_indent()

            out += f'{self.indent}std::string str(size_of_head(), \'\\0\');\n'
            out += f'{self.indent}bragi::limited_writer wr{{str.data(), str.size()}};\n\n'
            out += self.emit_assert_that('encode_head(wr)') + '\n'
            out += f'{self.indent}return str;\n'
            self.leave_indent()

            out += f'{self.indent}}}\n\n'

        return out

    # Protobuf compatibilty code
    def emit_parse_from_array(self, parent):
        out = f'{self.indent}bool ParseFromArray(const void *data, size_t size) {{\n'
        self.enter_indent()

        out += f'{self.indent}bragi::limited_reader rd{{data, size}};\n\n'
        out += f'{self.indent}return decode_head(rd);\n'

        self.leave_indent()

        out += f'{self.indent}}}\n\n'

        return out

    def emit_accessors(self, members):
        out = ''

        for m in members:
            # getters
            ref = '&' if m.type.identity not in {TypeIdentity.INTEGER, TypeIdentity.CONSTS, TypeIdentity.ENUM} else ''

            out += f'{self.indent}{self.generate_type(m.type)} {ref}{m.name}() {{\n'
            self.enter_indent()
            out += f'{self.indent}return m_{m.name};\n'
            self.leave_indent()
            out += f'{self.indent}}}\n\n'

            if m.type.identity is TypeIdentity.ARRAY:
                ref = '&' if m.type.subtype.identity not in {TypeIdentity.INTEGER, TypeIdentity.CONSTS, TypeIdentity.ENUM} else ''

                out += f'{self.indent}{self.generate_type(m.type.subtype)} {ref}{m.name}(size_t i) {{\n'
                self.enter_indent()
                out += f'{self.indent}return m_{m.name}[i];\n'
                self.leave_indent()
                out += f'{self.indent}}}\n\n'

                out += f'{self.indent}size_t {m.name}_size() {{\n'
                self.enter_indent()
                out += f'{self.indent}return m_{m.name}.size();\n'
                self.leave_indent()
                out += f'{self.indent}}}\n\n'

            # setters
            out += f'{self.indent}void set_{m.name}({self.generate_type(m.type)} val) {{\n'
            self.enter_indent()
            out += f'{self.indent}p_{m.name} = true;\n'
            out += f'{self.indent}m_{m.name} = val;\n'
            self.leave_indent()
            out += f'{self.indent}}}\n\n'

            if m.type.identity is TypeIdentity.ARRAY:
                out += f'{self.indent}void set_{m.name}(size_t i, {self.generate_type(m.type.subtype)} val) {{\n'
                self.enter_indent()
                out += f'{self.indent}p_{m.name} = true;\n'
                out += f'{self.indent}m_{m.name}[i] = val;\n'
                self.leave_indent()
                out += f'{self.indent}}}\n\n'

                if m.type.n_elements is None:
                    out += f'{self.indent}void add_{m.name}({self.generate_type(m.type.subtype)} v) {{\n'
                    self.enter_indent()
                    out += f'{self.indent}p_{m.name} = true;\n'
                    out += f'{self.indent}m_{m.name}.push_back(v);\n'
                    self.leave_indent()
                    out += f'{self.indent}}}\n\n'

        return out

    def emit_constructor(self, name, members):
        out = f'{self.indent}{name}({self.stdlib_traits.allocator_argument()})'

        if len(members) > 0 or self.stdlib_traits.needs_allocator():
            out += f'\n{self.indent}: '
            for i, m in enumerate(members):
                alloc = self.stdlib_traits.allocator_parameter() if self.check_needs_allocator(m.type) else ''
                out += f'm_{m.name}{{{alloc}}}, p_{m.name}{{false}}'

                if i < len(members) - 1 or self.stdlib_traits.needs_allocator():
                    out += f', \n{self.indent}  '

        if self.stdlib_traits.needs_allocator():
            out += f'allocator{{{self.stdlib_traits.allocator_parameter()}}}'

        out += ' { }\n\n'

        return out

    def emit_class_members(self, members):
        out = ''

        if len(members):
            self.leave_indent()
            out += f'{self.indent}private:\n'
            self.enter_indent()
            for m in members:
                out += f'{self.indent}{self.generate_type(m.type)} m_{m.name}; bool p_{m.name};\n'

        if self.stdlib_traits.needs_allocator():
            out += f'{self.indent}Allocator allocator;\n'

        return out

    def generate_message(self, message):
        all_members = flatten([
            flatten((m.members if type(m) is TagsBlock else [m] for m in message.head.members) if message.head is not None else []),
            flatten((m.members if type(m) is TagsBlock else [m] for m in message.tail.members) if message.tail is not None else [])
        ])

        out = ''
        if self.stdlib_traits.needs_allocator():
            out += f'{self.indent}template <typename Allocator>\n'

        out += f'{self.indent}struct {message.name} {{\n'
        self.enter_indent()
        out += f'{self.indent}static constexpr uint32_t message_id = {message.id};\n'
        out += f'{self.indent}static constexpr bool has_tail = {"true" if message.tail else "false"};\n'
        out += f'{self.indent}static constexpr size_t head_size = {message.head.size};\n\n'

        out += self.emit_constructor(message.name, all_members)
        out += self.emit_accessors(all_members)

        if message.head:
            out += self.emit_calculate_size_of('head', message.head.members, message)
            out += self.emit_part_encoder('head', message, message.head.members)
            out += self.emit_part_decoder('head', message, message.head.members)
        else:
            out += self.emit_stub_calculate_size_of('head')
            out += self.emit_part_encoder('head', None, None)
            out += self.emit_part_decoder('head', None, None)
        if message.tail:
            out += self.emit_calculate_size_of('tail', message.tail.members, message)
            out += self.emit_part_encoder('tail', message, message.tail.members)
            out += self.emit_part_decoder('tail', message, message.tail.members)
        else:
            out += self.emit_stub_calculate_size_of('tail')
            out += self.emit_part_encoder('tail', None, None)
            out += self.emit_part_decoder('tail', None, None)

        if self.protobuf_compat:
            out += self.emit_serialize_as_string()
            out += self.emit_parse_from_array(message)

        out += self.emit_class_members(all_members)

        self.leave_indent()

        return out + f'{self.indent}}}; // struct {message.name}\n\n'

    def generate_struct(self, struct):
        all_members = flatten([
            flatten((m.members if type(m) is TagsBlock else [m] for m in struct.members))
        ])

        out = ''
        if self.stdlib_traits.needs_allocator():
            out += f'{self.indent}template <typename Allocator>\n'

        out += f'{self.indent}struct {struct.name} {{\n'
        self.enter_indent()

        out += self.emit_constructor(struct.name, all_members)
        out += self.emit_accessors(all_members)

        out += self.emit_struct_calculate_size_of(struct.members, struct)
        out += self.emit_struct_encoder(struct, struct.members)
        out += self.emit_struct_decoder(struct, struct.members)

        out += self.emit_class_members(all_members)

        self.leave_indent()

        return out + f'{self.indent}}}; // struct {struct.name}\n\n'
